"""Various functions for time series anlysis.
If not specified differently, references are given on Luetkepohl "New
Introduction to Multiple Time Series Analysis".

.. currentmodule:: varwg.time_series_analysis.models

.. autosummary::
   :nosignatures:
   :toctree: generated/

   VAR_LS
   VAR_LS_sim
   VAREX_LS
   VAREX_LS_sim
   VAR_order_selection
   VAR_residuals
   VAREX_residuals
   VAR_LS_predict
   AIC
   FPE
   HQ
   SC
"""

import itertools
from collections import namedtuple
from warnings import warn

import numpy as np
import scipy
from scipy import linalg, optimize
from scipy.linalg import kron
from scipy.stats import skew
from tqdm import tqdm

import varwg
from varwg import helpers as my
from varwg.time_series_analysis import time_series as ts
from varwg.time_series_analysis.distributions import MDFt
from varwg.time_series_analysis import phase_randomization


mgarch_param_factory = namedtuple(
    "mgarch_param_factory", ("gamma0", "Gammas", "Gs", "cov_residuals")
)


def MGARCH_ML(residuals, q, m):
    """Estimating MGARCH parameters by numerically maximizing the
    log-likelihood.

    """
    K, T = residuals.shape
    dim0 = int(0.5 * K * (K + 1))
    gamma0 = np.full(dim0, 1e-6)
    Gammas = q * [np.diagflat(np.full(dim0, 0.1))]
    Gs = m * [np.diagflat(np.full(dim0, 0.1))]
    delta_vec = vec(
        np.hstack((gamma0[:, None], np.hstack(Gammas), np.hstack(Gs)))
    )
    sigma0 = np.cov(residuals)
    result = optimize.minimize(
        _MGARCH_likelihood,
        x0=delta_vec,
        bounds=len(delta_vec) * [(0, None)],
        args=(residuals, q, m, sigma0),
        options=dict(disp=True),
    )
    gamma0, Gammas, Gs = _MGARCH_unpack(result.x, K, q, m)
    _check_stationarity(Gammas, Gs)
    mgarch_residuals = MGARCH_residuals(residuals, gamma0, Gammas, Gs)
    cov_residuals = np.cov(mgarch_residuals)
    params = mgarch_param_factory(gamma0, Gammas, Gs, cov_residuals)
    return params


def _check_stationarity(Gammas, Gs):
    M = np.zeros_like(Gammas[0])
    for array in itertools.chain(Gammas, Gs):
        M += array
    eigenvalues = linalg.eigvals(M)
    if np.any(eigenvalues >= 1):
        msg = "Non-stationarity: MGARCH eigenvalues too high. "
        msg = msg + "\n" + repr(eigenvalues)
        raise ValueError(msg)


def _MGARCH_likelihood(params, ut, q, m, sigma0):
    K, T = ut.shape
    gamma0, Gammas, Gs = _MGARCH_unpack(params, K, q, m)
    sigmas = m * [sigma0]
    llh = -K * np.log(2 * np.pi) * T

    def llh_single(t):
        llh = 0
        sigma_vech = gamma0
        for j in range(q):
            sigma_vech += Gammas[j] * vech(ut[:, t - j] * ut[:, t - j].T)
        for j in range(m):
            sigma_vech += Gs[j] * vech(sigmas[m - j - 1])
        sigma = unvech(sigma_vech, K)
        sigmas[:-1] = sigmas[1:]
        sigmas[-1] = sigma

        # punishment for mathematically unsound behaviour
        variance_min = np.diagonal(sigma).min()
        try:
            sigma_det = linalg.det(sigma)
            # print(sigma_det)
        except ValueError:
            sigma_det = -np.inf

        if variance_min <= 0:
            raise OverflowError
            # import ipdb; ipdb.set_trace()
            llh += variance_min**2
        elif sigma_det <= 0:
            llh += sigma_det**2
        elif not np.isfinite(sigma_det):
            llh += 1e9
        else:
            llh += -np.log(sigma_det) - np.dot(
                np.dot(ut[:, t].T, linalg.inv(sigma)), ut[:, t]
            )
        return llh

    for t in range(max(q, m), T):
        try:
            llh_t = llh_single(t)
        except OverflowError:
            # this likelihood is bad enough, no need to go on...
            break
        llh += llh_t

    # llh *= .5
    return -llh


def MGARCH_residuals(ut, gamma0, Gammas, Gs):
    K, T = ut.shape
    residuals = np.zeros((K, T), dtype=float)
    q, m = len(Gammas), len(Gs)
    sigmas = m * [np.cov(ut)]
    for t in range(max(q, m), T):
        sigma_vech = gamma0
        for j in range(q):
            sigma_vech += Gammas[j] * vech(ut[:, t - j] * ut[:, t - j].T)
        for j in range(m):
            sigma_vech += Gs[j] * vech(sigmas[m - j - 1])
        sigma = unvech(sigma_vech, K)
        sigmas[:-1] = sigmas[1:]
        sigmas[-1] = sigma
        # residuals[:, t] = np.squeeze(sigma.dot(sigma) * ut[:, t])
        residuals[:, t] = np.squeeze(linalg.inv(sqrtm(sigma)) * ut[:, t])
    return residuals


def sqrtm(A):
    """Square root of a positive-definite Matrix.

    >>> A = np.array([[4., 0, 0], [0, 9., 0], [0, 0, 16.]])
    >>> sqrtm(A)
    array([[2., 0., 0.],
           [0., 3., 0.],
           [0., 0., 4.]])

    Notes
    -----
    See Appendix 9.4.
    """
    eigenvalues, P = linalg.eigh(A)
    Lambda_sqrt = np.diag(np.sqrt(eigenvalues))
    return P @ Lambda_sqrt @ P.T


def MGARCH_sim(params, T, sigma0, epsilon=None, n_presim_steps=100):
    gamma0, Gammas, Gs, cov_residuals = params
    q, m = len(Gammas), len(Gs)
    K = cov_residuals.shape[0]
    n_sim_steps = T + max(q, m) + n_presim_steps
    if epsilon is None:
        # epsilon = np.random.multivariate_normal(K * [0], cov_residuals,
        #                                         n_sim_steps - max(q, m))
        epsilon = varwg.rng.multivariate_normal(
            K * [0], cov_residuals, n_sim_steps - max(q, m)
        )
        epsilon = epsilon.T
    ut = np.zeros((K, n_sim_steps))
    ut[:, : max(q, m)] = epsilon[:, : max(q, m)]
    sigmas = q * [sigma0]
    for t in range(max(q, m), n_sim_steps):
        sigma_vech = gamma0
        for j in range(q):
            sigma_vech += Gammas[j] @ vech(ut[:, t - j] @ ut[:, t - j].T)
        for j in range(m):
            sigma_vech += Gs[j] @ vech(sigmas[m - j - 1])
        sigma = unvech(sigma_vech, K)
        sigmas[:-1] = sigmas[1:]
        sigmas[-1] = sigma
        ut[:, t] = linalg.sqrtm(sigma) @ epsilon[t]
    return ut[:, -T:]


def _MGARCH_unpack(params, K, q, m):
    dim0 = int(0.5 * K * (K + 1))
    delta = unvec(params, dim0)
    gamma0 = delta[:, 0]
    Gammas = [delta[:, i : i + dim0] for i in range(1, 1 + q * dim0, dim0)]
    Gs = [
        delta[:, i : i + dim0]
        for i in range(1 + q * dim0, 1 + (q + m) * dim0, dim0)
    ]
    return gamma0, Gammas, Gs


def VAR_LS(data, p=2, biased=True):
    """Least-Squares parameter estimation for a vector auto-regressive model of
    the form Y = B*Z + U. Records containing nans are excluded.
    Refer to the Least-Squares Estimator example 3.2.3 p.78. for method and
    variable names.

    Parameters
    ----------
    data : (K, T) ndarray
        K is the number of variables, T the number of timesteps
    p : int
        Autoregressive order of the process.

    Returns
    -------
    B : array
        Parameters of the fitted VAR-process.
    sigma_u : array
        Covariance matrix of the residuals.
    biased : bool, optional
        If true, use the number of non-nan observations (n_obs) to
        'unbias' sigma_u. Otherwise, use n_obs - K * p - 1.

    See also
    --------
    VAR_order_selection : Helps to find a p for parsimonious estimation.
    VAR_residuals : Returns the residuals based on given data and LS estimator
    VAR_LS_sim : Simulation based on LS estimator.
    VAR_LS_predict : Predict given prior data and LS estimator.

    """
    # number of variables
    if np.ndim(data) == 2:
        K, T = data.shape[0], data.shape[1] - p
    elif np.ndim(data) == 1:
        K, T = 1, len(data) - p
        data = data[np.newaxis, :]
    # Y is a (K, T - p) array
    Y = data[:, p:]

    Z = np.empty((K * p + 1, T))
    Zt = np.empty(K * p + 1)
    Zt[0] = 1
    for t in range(p, T + p):
        for subt in range(p):
            start_i = 1 + subt * K
            stop_i = 1 + (subt + 1) * K
            Zt[start_i:stop_i] = data[:, t - subt - 1]
        Z[:, t - p] = Zt

    # delete all columns containing nans
    mask = ~np.isnan(Y).any(axis=0) & ~np.isnan(Z).any(axis=0)
    Y = Y[:, mask]
    Z = Z[:, mask]

    if Y.shape[1] <= K * p + 1:
        warn("High number of nans. Doing ridge regularization.")
        inv = np.linalg.inv(Z @ Z.T + 1e-6 * np.eye(Z.shape[0]))
    else:
        inv = np.linalg.inv(Z @ Z.T)
    B = Y @ Z.T @ inv

    # covariance matrix of the noise
    U = Y - B @ Z
    if biased:
        denominator = Y.shape[1]
    else:
        denominator = Y.shape[1] - K * p - 1
    sigma_u = U @ U.T / denominator

    return B, sigma_u


def VAREX_LS(data, p, ex):
    """Least-Squares parameter estimation for a vector auto-regressive model of
    the form

    ..math::y_t = A_1 y_{t-1} + ... + A_p y_{t-p} + C x_t + u_t

    Records containing nans are excluded.
    Refer to the Least-Squares Estimator example 3.2.3 p.78. for method and
    variable names.

    Parameters
    ----------
    data : (K, T) ndarray
        K is the number of variables, T the number of timesteps
    ex : (T,) ndarray
        An external variable
    p : int
        Autoregressive order of the process.

    Returns
    -------
    B : array
        Parameters of the fitted VAR-process.
        B := (A_1, ..., A_p, C)
    sigma_u: array
        Covariance matrix of the residuals of the data.

    See also
    --------
    VAR_order_selection : Helps to find a p for parsimonious estimation.
    VAR_residuals : Returns the residuals based on given data and LS estimator
    VAR_LS_sim : Simulation based on LS estimator.
    VAR_LS_predict : Predict given prior data and LS estimator.
    """
    # number of variables
    if np.ndim(data) == 2:
        K, T = data.shape[0], data.shape[1] - p
    elif np.ndim(data) == 1:
        K, T = 1, len(data) - p
        data = data[np.newaxis, :]
    # Y is a (K, T) array
    Y = data[:, p:]

    Z = np.empty((K * p + 1, T))
    Zt = np.empty((K * p, 1))
    Z[-1] = ex[p:].reshape(1, T)
    for t in range(p, T + p):
        for subt in range(p):
            start_i = subt * K
            stop_i = (subt + 1) * K
            Zt[start_i:stop_i] = data[:, t - subt - 1].reshape((K, 1))
        Z[:-1, t - p] = Zt

    # delete all columns containing nans
    Y_nan_cols = np.where(np.isnan(Y))[1]
    Y = np.delete(Y, Y_nan_cols, axis=1)
    Z = np.delete(Z, Y_nan_cols, axis=1)
    Z_nan_cols = np.where(np.isnan(Z))[1]
    Y = np.delete(Y, Z_nan_cols, axis=1)
    Z = np.delete(Z, Z_nan_cols, axis=1)

    # B contains all the parameters we want: (A1, ..., Ap, C)
    # Y = BZ + U
    B = Y @ Z.T @ np.linalg.inv(Z @ Z.T)

    # covariance matrix of the noise of data
    sigma_u = Y @ Y.T - B @ Z @ Y.T
    sigma_u /= Y.shape[1] - K * p - 1

    return B, sigma_u


def SVAR_LS(
    data, doys, p=2, doy_width=60, fft_order=3, var_names=None, verbose=True
):
    """Seasonal version of the least squares estimator."""
    K, T = data.shape
    Bs, sigma_us = [], []
    unique_doys = np.unique(doys)
    for doy in tqdm(unique_doys, disable=(not verbose)):
        mask = (doys > doy - doy_width) & (doys <= doy + doy_width)
        if (doy - doy_width) < 0:
            mask |= doys > (365.0 - doy_width + doy)
        if (doy + doy_width) > 365:
            mask |= doys < (doy + doy_width - 365.0)
        B, sigma_u = VAR_LS(np.where(mask, data, np.nan), p=p)
        Bs += [B]
        sigma_us += [sigma_u]

    Bs, sigma_us = np.asarray(Bs), np.asarray(sigma_us)

    def matr_fft(M, fft_order):
        smoothed = np.zeros((M.shape[1], M.shape[2], M.shape[0]))
        for k in range(K):
            for j in range(M.shape[2]):
                smoothed[k, j] = my.fourier_approx(
                    my.interp_nonfin(M[:, k, j], fft_order)
                )
        return smoothed

    def cov_matr_fft(M, fft_order):
        smoothed = np.zeros_like(M)
        M_log = np.asarray([matrix_log_spd(cov).real for cov in M])
        for k in range(K):
            for j in range(M.shape[2]):
                smoothed[:, k, j] = my.fourier_approx(
                    my.interp_nonfin(M_log[:, k, j], fft_order)
                )
        smoothed = np.asarray([matrix_exp_sym(cov.real) for cov in smoothed])
        return np.moveaxis(smoothed, 0, -1)

    Bs = matr_fft(Bs, fft_order)
    sigma_us = cov_matr_fft(sigma_us, fft_order)
    return Bs, sigma_us


def VAR_mean(B):
    K = B.shape[0]
    p = (B.shape[1] - 1) // K
    Ai = [np.asarray(B[:, 1 + i * K : 1 + (i + 1) * K]) for i in range(p)]
    nu = B[:, 0]
    mu = np.identity(K)
    for i in range(p):
        mu -= Ai[i]
    return (np.linalg.inv(mu) @ nu.T)[:, None]


# def VAR_cov(B, sigma_u):
#     """see p. 27-28"""
#     K = B.shape[0]
#     p = (B.shape[1] - 1) // K
#     A = B2A(B)
#     Ikp2 = np.identity((K * p) ** 2)
#     sigma_U = np.zeros_like(A)
#     sigma_U[:K, :K] = sigma_u
#     cov_vec = np.linalg.inv(Ikp2 - np.kron(A, A)) @ vec(sigma_U)
#     return unvec(cov_vec, K)[:K, :K]


def B2A(B):
    K = B.shape[0]
    p = (B.shape[1] - 1) // K
    A = np.zeros((K * p, K * p))
    A[:K] = B[:, 1:]
    Ik = np.identity(K)
    for i in range(p - 1):
        A[(i + 1) * K : (i + 2) * K, i * K : (i + 1) * K] = Ik
    return A


@my.cache("ut")
def SVAR_LS_sim(
    Bs,
    sigma_us,
    doys,
    m=None,
    ia=None,
    m_trend=None,
    u=None,
    n_presim_steps=100,
    fixed_data=None,
    phase_randomize=False,
    rphases=None,
    return_rphases=False,
    p_kwds=None,
    taboo_period_min=None,
    taboo_period_max=None,
    verbose=False,
):
    if p_kwds is None:
        p_kwds = dict()
    doys_ii = (doys % 365) / 365.0 * len(np.unique(doys))
    # doys_ii = doys % len(np.unique(doys))
    doys_ii = doys_ii.astype(int)
    K = Bs.shape[0]
    p = (Bs.shape[1] - 1) // K
    T = len(doys)
    Y = np.zeros((K, T + p))
    Y[:, :p] = VAR_mean(Bs[..., doys_ii[-1]])
    if phase_randomize:
        if u is None:
            raise RuntimeError("u must be passed for phase randomization!")
        if verbose:
            print("Phase randomizing residuals")
        u = phase_randomization.randomize2d(
            u,
            T=T,
            taboo_period_min=taboo_period_min,
            taboo_period_max=taboo_period_max,
            return_rphases=return_rphases,
            rphases=rphases,
            **p_kwds,
        )
        if return_rphases:
            u, rphases = u
    if u is None:
        u = np.array(
            [
                varwg.rng.multivariate_normal(K * [0], sigma_us[..., doy_i])
                for doy_i in doys_ii
            ]
        )
        u = u.T
    SVAR_LS_sim.ut = u
    for t, doy_i in enumerate(doys_ii):
        Y[:, t + p] = VAR_LS_sim(
            Bs[..., doy_i],
            # ignored when using phase_randomization!
            sigma_us[..., doy_i],
            1,
            None if m is None else m[:, t, None],
            None if ia is None else ia[:, t, None],
            m_trend,
            n_presim_steps=0,
            # u=None if u is None else u[:, t, None],
            u=u[:, t, None],
            prev_data=Y[:, t : t + p],
            # u is already phase randomized above if requested
            phase_randomize=False,
        ).ravel()
    Y = Y[:, p:]
    Y_new = Y

    # # SVAR messes with the marginals, so we z-transform to normal as
    # # expected by the rest of the code base. We maintain the means,
    # # though.
    # means = np.array([np.squeeze(VAR_mean(Bs[..., doys_ii[t]]))
    #                   for t in range(T)]).T
    # stds = np.array([np.sqrt(np.diag(
    #     VAR_cov(Bs[..., doys_ii[t]],
    #             sigma_us[..., doys_ii[t]])))
    #                  for t in range(T)]).T
    # means_mean = means.mean(axis=-1)[:, None]
    # stds_mean = stds.mean(axis=-1)[:, None]
    # Y_new = (Y - means_mean) / stds_mean * stds + means

    # import matplotlib.pyplot as plt
    # fig, axs = plt.subplots(nrows=K, ncols=2,
    #                         sharex=True, sharey="row",
    #                         constrained_layout=True)
    # for var_i in range(K):
    #     axs[var_i, 0].plot(Y[var_i])
    #     axs[var_i, 0].plot(means[var_i])
    #     axs[var_i, 0].plot(means[var_i] - stds[var_i], "--k")
    #     axs[var_i, 0].plot(means[var_i] + stds[var_i], "--k")
    #     axs[var_i, 0].axhline(means_mean[var_i])
    #     axs[var_i, 0].axhline(means_mean[var_i] - stds_mean[var_i],
    #                           linestyle="--", color="b")
    #     axs[var_i, 0].axhline(means_mean[var_i] + stds_mean[var_i],
    #                           linestyle="--", color="b")
    #     axs[var_i, 1].plot(Y_new[var_i], alpha=.25, label="new")
    #     axs[var_i, 1].plot(Y[var_i], alpha=.25, label="old")
    # axs[0, 1].legend(loc="best")
    # plt.show()

    if return_rphases:
        return Y_new, rphases
    return Y_new


@my.cache("ut")
def SVAR_LS_fill(
    Bs,
    sigma_us,
    doys,
    Y,
    A=None,
    m=None,
    ia=None,
    m_trend=None,
    n_presim_steps=100,
    fixed_data=None,
    p_kwds=None,
):
    if p_kwds is None:
        p_kwds = dict()
    doys_ii = (doys % 365) / 365.0 * len(np.unique(doys))
    doys_ii = doys_ii.astype(int)
    K = Bs.shape[0]
    p = (Bs.shape[1] - 1) // K
    Y = np.hstack((np.zeros((K, p)), Y))
    Y[:, :p] = VAR_mean(Bs[..., doys_ii[-1]])
    for t, doy_i in enumerate(doys_ii):
        if np.any(np.isnan(Y[:, t + p])):
            Y_p = VAR_LS_sim(
                Bs[..., doy_i],
                # ignored when using phase_randomization!
                sigma_us[..., doy_i],
                1,
                None if m is None else m[:, t, None],
                None if ia is None else ia[:, t, None],
                m_trend,
                n_presim_steps=0,
                u=np.zeros(K)[:, None],
                prev_data=Y[:, t : t + p],
                # u is already phase randomized above if requested
                phase_randomize=False,
            ).ravel()
            u_t = Y[:, t + p] - Y_p
            u_t = _cholesky_partial(u_t, sigma_us[..., doy_i], A=A)
            Y[:, t + p] = Y_p + u_t
    return Y[:, p:]


def scale_z(z, i, x_i, A):
    z = np.copy(z)
    z[i] = (x_i - np.sum(A[i, :i] * z[:i])) / A[i, i]
    return z


def _cholesky_partial(u, sigma_us, A=None):
    if A is None:
        try:
            A = np.linalg.cholesky(sigma_us)
        except np.linalg.LinAlgError:
            print("Adding a little random noise in order to invert matrix...")
            sigma_us += (
                varwg.rng.normal(size=sigma_us.shape) * sigma_us.std() * 1e-9
            )
            A = np.linalg.cholesky(sigma_us)
    finite_mask = np.isfinite(u)
    z = np.empty_like(u)
    z[~finite_mask] = varwg.rng.normal(size=np.sum(~finite_mask))
    z[finite_mask] = u[finite_mask]
    ii = np.where(finite_mask)[0]
    for i, u_i in zip(ii, u[finite_mask]):
        z = scale_z(z, i, u_i, A)
    z = A @ z
    # assert np.all(np.isclose(z[finite_mask], u[finite_mask]))
    return z


def VAR_LS_asy(data, skewed_i, p=None):
    """Skewness of the differences of the residuals."""
    if p is None:
        p = VAR_order_selection(data)
    B, _ = VAR_LS(data, p)
    residuals = VAR_residuals(data, B, p)
    if isinstance(skewed_i, int):
        skewed_i = [skewed_i]
    asymmetries = []
    for i in skewed_i:
        diff = np.diff(residuals[i])
        asymmetries += [skew(diff[np.isfinite(diff)])]
    return asymmetries


@my.cache("ut")
def VAR_LS_sim_asy(
    B,
    sigma_u,
    T,
    data,
    p,
    skewed_i,
    n_presim_steps=100,
    verbose=False,
    var_names=None,
    *args,
    **kwds,
):
    residuals = VAR_residuals(data, B, p)
    data_skew = VAR_LS_asy(data, skewed_i, p)
    data_skew = np.squeeze(data_skew)
    K = data.shape[0]
    n_sim_steps = T + p + n_presim_steps
    # u = np.random.multivariate_normal(K * [0], sigma_u, n_sim_steps - p).T
    params = MDFt.fit(residuals)
    u = MDFt.sample(n_sim_steps, *params)
    u_initial = np.copy(u)

    VAR_LS_sim_asy.ut = u

    fig = my.splom(residuals)
    fig.suptitle("Residuals")
    fig = my.splom(u)
    fig.suptitle("Simulated residuals")

    if var_names is None:
        var_names = [str(i) for i in range(K)]

    def asy(u):
        return skew(np.diff(u[skewed_i], axis=1), axis=1)

    def skew_error(u):
        u_skew = asy(u)
        return np.sum((data_skew - u_skew) ** 2)

    @my.cache("i", "j", "k", "di", "dj", "dk")
    def swap(data):
        # i, j, k = np.random.randint(data.shape[1], size=3)
        width = 10
        j = varwg.rng.randint(1, data.shape[1] - 2)
        i = varwg.rng.randint(max(0, j - 2), j)
        k = varwg.rng.randint(j + 1, min(data.shape[1] - 1, j + width))
        di, dj, dk = data[skewed_i, [i, j, k]]
        if di < dj < dk:
            data[:, [j, k]] = dk, dj
        elif dk < di < dj:
            data[:, [i, k]] = dk, di
        elif dk < dj < di:
            data[:, [i, j, k]] = dk, di, dj
        elif dj < di < dk:
            data[:, [i, j, k]] = dj, dk, di
        swap.i, swap.j, swap.k = i, j, k
        swap.di, swap.dj, swap.dk = di, dj, dk
        return data

    def swap_back(data):
        data[:, [swap.i, swap.j, swap.k]] = swap.di, swap.dj, swap.dk
        return data

    # @my.cache("i", "j")
    # def swap(data):
    #     i, j = np.random.randint(data.shape[1], size=2)
    #     data[:, [i, j]] = data[:, [j, i]]
    #     swap.i, swap.j = i, j
    #     return data

    # def swap_back(data):
    #     data[:, [swap.j, swap.i]] = data[:, [swap.i, swap.j]]
    #     return data

    temp = 2.0
    k = 0.95
    m, M = 0, 200
    error_old = skew_error(u)
    u = swap(u)
    error_new = skew_error(u)
    error_best = min(error_old, error_new)
    error_last_best = error_best
    u_best = np.copy(u)

    if verbose:
        print("Data asymmetry: %.3f" % data_skew)
        print("Initial sim asymmetry: %.3f" % asy(u))

    # while iteration < 100 or abs(error_new - error_old) / error_old > 1e-9:
    iteration = 0
    accept = np.empty(M, dtype=bool)
    while True:
        iteration += 1
        accept[m] = True
        if error_new > error_old:
            p = np.exp((error_old - error_new) / temp)
            if not varwg.rng.rand() < p:
                accept[m] = False

        if error_new < error_best:
            error_last_best = error_best
            error_best = error_new
            u_best = np.copy(u)

        if accept[m]:
            error_old = error_new
            u = swap(u)
            error_new = skew_error(u)
        else:
            u = swap_back(u)

        m += 1
        if m == M:
            m = 0
            temp *= k
            print(
                "sim: %.3f, data: %.3f, temp: %.5f" % (asy(u), data_skew, temp)
            )
            if ((error_last_best - error_best) / error_last_best) < 1e-6 or (
                np.all(~accept) and iteration > 5000
            ):
                break

    import matplotlib.pyplot as plt

    fig, axs = plt.subplots(3, sharex=True, sharey=True)
    axs[0].plot(residuals[0], label="residuals")
    axs[0].plot(u_best[0], label="u")
    axs[0].legend(loc="best")
    axs[0].set_title("theta")

    axs[1].plot(residuals[1], label="residuals")
    axs[1].plot(u_best[1], label="u")
    axs[1].legend(loc="best")
    axs[1].set_title("U")

    axs[2].plot(residuals[2], label="residuals")
    axs[2].plot(u_best[2], label="u")
    axs[2].legend(loc="best")
    axs[2].set_title("wf")
    fig.suptitle("Residuals")

    for var_i in range(K):
        fig, axs = plt.subplots(ncols=2, sharex=True, sharey=True)
        both = [
            residuals[var_i],
            # scipy.stats.trimboth(residuals[var_i], .05),
            np.squeeze(u_best[var_i]),
        ]
        labels = "measured", "simulated"
        for ax, vals, label in zip(axs, both, labels):
            my.hist(
                vals,
                20,
                kde=True,
                dist=(
                    scipy.stats.distributions.norm,
                    scipy.stats.distributions.t,
                ),
                ax=ax,
            )
            skew_ = scipy.stats.skew(vals)
            kurt = scipy.stats.kurtosis(vals)
            ax.set_title(
                r"%s residuals $\gamma$=%.3f kurt=%.3f" % (label, skew_, kurt)
            )
        fig.suptitle(var_names[var_i])

    VAR_LS_sim_asy.ut = u_best
    return VAR_LS_sim(B, sigma_u, T, u=u_best, *args, **kwds)


def VAR_LS_sim(
    B,
    sigma_u,
    T,
    m=None,
    ia=None,
    m_trend=None,
    u=None,
    n_presim_steps=100,
    fixed_data=None,
    prev_data=None,
    transform=None,
    phase_randomize=False,
    rphases=None,
    return_rphases=False,
    p_kwds=None,
    taboo_period_min=None,
    taboo_period_max=None,
):
    """Based on a least squares estimator, simulate a time-series of the form
    ..math::y(t) = nu + A1*y(t-1) + ... + Ap*y(t-p) + ut
    B contains (nu, A1, ..., Ap).
    See p. 707f

    Parameters
    ----------
    B : (K, K*p+1) ndarray
        Parameters of the VAR-process as returned from VAR_LS. K is the number
        of variables, p the autoregressive order.
    sigma_u : (K, K) ndarray
        Covariance matrix of the residuals as returned from VAR_LS.
    T : int
        Number of timesteps to simulate.
    m : (K,) ndarray, optional
        Process means (will be scaled according to B).
    ia : (K, T) ndarray, optional
        Interannual variability. Additional time-varying disturbance to the
        process means (will be scaled according to B).
    m_trend : (K,) ndarray, optional
        Change in means, that will be applied linearly so that this change is
        reached after the T timesteps.
    u : (K, T) ndarray, optional
        Residuals to be used instead of multivariate gaussian serially
        independent random numbers.
    n_presim_steps : int, optional
        Number of presimulation timesteps that will be thrown away.
    fixed_data : (K, T) ndarray, optional
        Data that will be fixed, i.e. at each timestep, these will be put in
        instead of the actual simulated values. Where fixed_data is nan, the
        simulated values will not be overwritten.
    transform : callable, optional
        On-the-fly transformation that accepts a sequence of length K
        (values of all variables at one time step) and the time index t.

    Returns
    -------
    Y : (K, T) ndarray
        Simulated values.

    See also
    --------
    VAR_LS : Least-squares estimator (to get B and sigma_u).
    VAR_order_selection : Helps to find a p for parsimonious estimation.
    VAR_residuals : Returns the residuals based on given data and LS estimator
    VAR_LS_predict : Predict given prior data and LS estimator.
    """
    if p_kwds is None:
        p_kwds = dict()
    # number of variables
    K = B.shape[0]
    # order of VAR_LS-process
    p = (B.shape[1] - 1) // K
    n_sim_steps = T + p + n_presim_steps

    if m is not None:
        m = _scale_additive(m, B[:, 1:], p)
        # we have to expand m to include the pre-simulation timesteps
        m = np.concatenate((m[:, :n_presim_steps], m), axis=1)

    if ia is not None:
        ia = _scale_additive(ia, B[:, 1:], p)

    # the first p columns are initial values, which will be omitted later
    Y = np.zeros((K, n_sim_steps))
    if prev_data is not None:
        Y[:, :p] = prev_data[:, -p:]
        n_sim_steps -= n_presim_steps

    Ai = [np.asarray(B[:, 1 + i * K : 1 + (i + 1) * K]) for i in range(p)]

    if m is None and prev_data is None:
        # setting starting values to the process mean
        Y[:, :p] = VAR_mean(B)
        Y = np.asarray(Y)

    if phase_randomize:
        if u is None:
            raise RuntimeError("u must be passed for phase randomization!")
        u = phase_randomization.randomize2d(
            u,
            T=T,
            # taboo_period_min=taboo_period_min,
            # taboo_period_max=taboo_period_max,
            return_rphases=return_rphases,
            rphases=rphases,
            **p_kwds,
        )
        if return_rphases:
            u, rphases = u
    elif u is None:
        u = varwg.rng.multivariate_normal(K * [0], sigma_u, n_sim_steps - p)
        u = u.T

    Y[:, -u.shape[1] :] += u
    if m is None:
        nu = B[:, 0]
        Y[:, p:] += nu[:, None]
    elif m is not None:
        Y[:, -m.shape[1] :] += m

    if ia is not None:
        Y[:, -ia.shape[1] :] += ia

    if m_trend is not None:
        # apply changes as a trend
        m_trend = np.asarray(m_trend)[:, None]
        m_trend = _scale_additive(m_trend, B[:, 1:], p)
        Y[:, -T:] += np.arange(T, dtype=float) / T * m_trend

    start_t = n_sim_steps - T
    for t in range(p, n_sim_steps):
        for i in range(p):
            Y[:, t] += Ai[i] @ Y[:, t - i - 1]

        # on-line transformations
        if transform:
            Y[:, t] = transform(Y[:, t], t - start_t)

        if (fixed_data is not None) and (t >= start_t):
            # fixing what's asked to be held constant
            Y[:, t] = np.where(
                np.isnan(fixed_data[:, t - start_t]),
                Y[:, t],
                fixed_data[:, t - start_t],
            )
    if return_rphases:
        return Y[:, -T:], rphases
    return Y[:, -T:]


def VAREX_LS_sim(
    B,
    sigma_u,
    T,
    ex,
    m=None,
    ia=None,
    m_trend=None,
    u=None,
    n_presim_steps=100,
    prev_data=None,
    ex_kwds=None,
):
    """Based on a least squares estimator, simulate a time-series of the form
    ..math::y(t) = A1*y(t-1) + ... + Ap*y(t-p) + C*x(t-1) + ut
    B contains (A1, ..., Ap, C).
    See p. 707f

    Parameters
    ----------
    B : (K, K*p+1) ndarray
        Parameters of the VAR-process as returned from VAR_LS. K is the number
        of variables, p the autoregressive order.
    sigma_u : (K, K) ndarray
        Covariance matrix of the residuals as returned from VAR_LS.
    ex : (T,) ndarray or function
        External variable. If given as a function, ex_t will be generated by
        calling ex(Y[:t], **ex_kwds), with Y being the simulated values.
    T : int
        Number of timesteps to simulate.
    m : (K,) ndarray, optional
        Process means (will be scaled according to B).
    ia : (K, T) ndarray, optional
        Interannual variability. Additional time-varying disturbance to the
        process means (will be scaled according to B).
    m_trend : (K,) ndarray, optional
        Change in means, that will be applied linearly so that this change is
        reached after the T timesteps.
    u : (K, T) ndarray, optional
        Residuals to be used instead of multivariate gaussian serially
        independent random numbers.
    n_presim_steps : int, optional
        Number of presimulation timesteps that will be thrown away.
    ex_kwds : dict, optional
        Keyword arguments to be passed to ex.

    Returns
    -------
    Y : (K, T) ndarray
        Simulated values.
    ex_out : (T,) ndarray
        External variable.

    See also
    --------
    VAR_LS : Least-squares estimator (to get B and sigma_u).
    VAR_order_selection : Helps to find a p for parsimonious estimation.
    VAR_residuals : Returns the residuals based on given data and LS estimator
    VAR_LS_predict : Predict given prior data and LS estimator.
    """
    # number of variables
    K = B.shape[0]
    # order of VAR_LS-process
    p = (B.shape[1] - 1) // K
    n_sim_steps = T + p + n_presim_steps

    try:
        len(ex)
        ex_isfunc = False
        ex_out = ex
    except TypeError:
        ex_kwds = {} if ex_kwds is None else ex_kwds
        ex_isfunc = True
        ex_out = np.empty(T)

    if m is not None:
        m = _scale_additive(m, B[:, 1:], p)
        # we have to expand m to include the pre-simulation timesteps
        m = np.concatenate((m[:, :n_presim_steps], m), axis=1)
    m_trend = np.asarray([0] * K) if m_trend is None else np.asarray(m_trend)
    m_trend = m_trend[:, np.newaxis]  # ha! erwischt!
    m_trend = _scale_additive(m_trend, B[:, 1:], p)

    if ia is not None:
        ia = _scale_additive(ia, B[:, 1:], p)

    # the first p columns are initial values, which will be omitted later
    Y = np.zeros((K, n_sim_steps))
    if prev_data is not None:
        Y[:, :p] = prev_data[:, -p:]
        n_sim_steps -= n_presim_steps

    Ai = [B[:, i * K : (i + 1) * K] for i in range(p)]
    C = B[:, -1]

    if u is None:
        u = varwg.rng.multivariate_normal(K * [0], sigma_u, n_sim_steps - p)
        u = u.T

    Y[:, -u.shape[1] :] += u
    if m is not None:
        Y[:, -m.shape[1] :] += m

    if ia is not None:
        Y[:, -ia.shape[1] :] += ia

    # apply changes as a trend
    Y[:, -T:] += np.arange(T, dtype=float) / T * m_trend

    start_t = n_sim_steps - T
    for t in range(p, n_sim_steps):
        for i in range(p):
            Y[:, t] += Ai[i] @ Y[:, t - i - 1]

        if t >= start_t:
            if ex_isfunc:
                ex_t = ex(Y[:, :t], **ex_kwds)
                ex_out[t - p - start_t] = ex_t
            else:
                ex_t = ex[t - p - start_t]
            Y[:, t] += np.squeeze(C * ex_t)

    return Y[:, -T:], ex_out


def VAR_residuals(data, B, p=2):
    K, T = data.shape

    # were we given a B with the nus (in the first column)?
    if B.shape[1] % (K * p) == 1:
        mean_adjusted = False
        i_shift = 1
        # what is the process mean?
        nu = B[:, 0]
        mu = VAR_mean(B)
        nu = np.asarray(nu).ravel()
    else:
        mean_adjusted = True
        i_shift = 0
        # estimate the process means from the data.
        mu = np.mean(data, axis=1)[:, None]

    # set the pre-sample period to the process means
    data = np.concatenate((np.empty((K, p)), data), axis=1)
    data[:, :p] = mu
    resi = np.copy(data)

    for t in range(p, T + p):
        for i in range(p):
            Ai = B[:, i_shift + i * K : i_shift + (i + 1) * K]
            resi[:, t] -= Ai @ data[:, t - i - 1]
        if not mean_adjusted:
            resi[:, t] -= nu
    return resi[:, p:] - mu if mean_adjusted else resi[:, p:]


def VAR_LS_extro(data, data_trans, transforms, backtransforms, p=2):
    B0, sigma_u = VAR_LS(data, p=p)
    B = np.copy(B0)
    K, T = data.shape
    M = B.shape[1]

    def errors(*Bs):
        B[:, 1:] = np.array(Bs).reshape((K, M - 1))
        pred = VAR_onestep_predictions(data, B, p=p)
        pred_trans = transforms(pred)
        return data_trans - pred_trans

    def error_sum(*Bs):
        return np.sum(errors(*Bs) ** 2)

    B_part = optimize.minimize(
        error_sum,
        x0=np.ravel(B0[:, 1:]),
        options=dict(disp=True),
        # method="Nelder-Mead",
        method="Powell",
    ).x
    B[:, 1:] = B_part.reshape((K, M - 1))
    residuals = VAR_LS_extro.residuals = errors(B[:, 1:])

    import matplotlib.pyplot as plt

    fig, axs = plt.subplots(nrows=K, sharex=True)
    pred = VAR_onestep_predictions(data, B, p=p)
    pred_trans = transforms(pred)
    for i, ax in enumerate(axs):
        ax.plot(data_trans[i], label="obs")
        ax.plot(pred_trans[i], label="est")
    axs[-1].legend()

    residuals_norm = backtransforms(residuals)
    finite_rows = np.all(np.isfinite(residuals_norm), axis=1)
    sigma_u = np.cov(residuals_norm[:, finite_rows])
    return B, sigma_u


def VAR_onestep_predictions(data, B, p=2):
    K, T = data.shape

    # were we given a B with the nus (in the first column)?
    if B.shape[1] % (K * p) == 1:
        mean_adjusted = False
        i_shift = 1
        # what is the process mean?
        nu = B[:, 0]
        mu = VAR_mean(B)
        nu = np.asarray(nu).ravel()
    else:
        mean_adjusted = True
        i_shift = 0
        # estimate the process means from the data.
        mu = np.mean(data, axis=1)[:, np.newaxis]

    predictions = np.zeros_like(data)
    predictions[:, :p] = mu
    # set the pre-sample period to the process means
    data = np.concatenate((np.empty((K, p)), data), axis=1)

    for t in range(p, T):
        for i in range(p):
            Ai = B[:, i_shift + i * K : i_shift + (i + 1) * K]
            predictions[:, t] += Ai @ data[:, t - i - 1]
        if not mean_adjusted:
            predictions[:, t] -= nu
    return np.asarray(predictions + mu) if mean_adjusted else predictions


def VAREX_residuals(data, ex, B, p=2, ex_kwds=None):
    K, T = data.shape

    try:
        len(ex)
        ex_isfunc = False
    except TypeError:
        ex_kwds = {} if ex_kwds is None else ex_kwds
        ex_isfunc = True

    # set the pre-sample period to the process means
    data = np.concatenate((np.zeros((K, p)), data), axis=1)
    resi = np.copy(data)

    C = np.asarray(B[:, -1])
    for t in range(p, T + p):
        for i in range(p):
            Ai = np.asarray(B[:, i * K : (i + 1) * K])
            resi[:, t] -= Ai @ data[:, t - i - 1]

        if ex_isfunc:
            ex_t = ex(data[:, :t], **ex_kwds)
        else:
            ex_t = ex[t - p]

        resi[:, t] -= np.squeeze(C * ex_t)
    return np.asarray(resi[:, p:])


def SVAR_residuals(data, doys, B, p=2):
    K, T = data.shape
    doys_ii = (doys % 365) / 365.0 * len(np.unique(doys))
    doys_ii = doys_ii.astype(int)

    # were we given a B with the nus (in the first column)?
    if B.shape[1] % (K * p) == 1:
        mean_adjusted = False
        i_shift = 1
    else:
        mean_adjusted = True
        i_shift = 0
        # estimate the process means from the data.
        mu = np.mean(data, axis=1)[:, np.newaxis]

    # set the pre-sample period to the process means
    data = np.concatenate((np.empty((K, p)), data), axis=1)
    data[:, :p] = VAR_mean(B[..., 0])
    resi = np.copy(data)

    for t in range(p, T + p):
        for i in range(p):
            Ai = B[:, i_shift + i * K : i_shift + (i + 1) * K, doys_ii[t - p]]
            resi[:, t] -= Ai @ data[:, t - i - 1]
        if not mean_adjusted:
            resi[:, t] -= B[:, 0, doys_ii[t - p]]
    resi = resi[:, p:] - mu if mean_adjusted else resi[:, p:]
    # A is smoothed by fft approximation. this can sometimes lead to
    # invalid matrices
    resi = my.interp_nonfin(resi)
    return resi


def VARMA_LS_prelim(data, p, q):
    """Preliminary version of the general simple least-squares vector
    autoregressive estimator for a vector autoregressive moving-average
    process. See p.474ff"""
    K, T = data.shape[0], data.shape[1] - p
    # number of parameters
    N = K**2 * (p + q)

    # first estimate ut by calculating the residuals from a long VAR-process
    B = VAR_LS(data, max(10, int(1.5 * (p + q))))[0]
    ut_est = VAR_residuals(data, B, p)

    Y = data[:, p:]
    X = np.ones((K * (p + q), T))
    Xt = np.zeros(K * (p + q))
    for t in range(p, T + p):
        for subt in range(p):
            start_i = subt * K
            stop_i = (subt + 1) * K
            Xt[start_i:stop_i] = data[:, t - subt - 1]
        for subt in range(p, p + q):
            start_i = subt * K
            stop_i = (subt + 1) * K
            Xt[start_i:stop_i] = ut_est[:, t - subt - p - 1]
        X[:, t - p] = Xt

    # R might not be necessary since we do not limit any parameters here
    R = np.identity(N)
    IK = np.identity(K)
    gamma = (
        (R.T @ np.linalg.inv(kron(X @ X.T, IK) @ R))
        @ R.T
        @ kron(X, IK)
        @ vec(Y)
    )
    residuals_arma_vec = vec(Y) - kron(X.T, IK) @ R @ gamma
    residuals_arma = residuals_arma_vec.reshape((K, T), order="F")
    # make the residuals have the same length as the data
    residuals_arma = np.concatenate((np.zeros((K, p)), residuals_arma), axis=1)
    sigma_u_arma = residuals_arma @ residuals_arma.T / T
    AM = gamma.reshape((K, -1), order="F")
    # the following expression leads to the same result...
    # AM = Y * X.T * (X * X.T).I
    return AM, sigma_u_arma, residuals_arma


def VARMA_LS_sim(
    AM,
    p,
    q,
    sigma_u,
    means,
    T,
    S=None,
    m=None,
    ia=None,
    m_trend=None,
    n_sim_multiple=2,
    fixed_data=None,
):
    """Generates a time series based on the VARMA-parameters AM.
    S and m should be sequences of length K. S is a variable-discerning
    multiplier and m a adder, respectively.

    Parameters
    ----------
    AM :       (K,K*(p+q)) array
               The parameters of the VARMA-process. The first p columns are
               interpreted as the Ai-matrices of the auto regressive part. The
               last q columns as the Mi-matrices of the moving average part. K
               is the number of variables simulated.
    p :        integer
               Order of the auto regressive process.
    q :        integer
               Order of the moving average process.
    sigma_u :  (K,K) array
               Covariance matrix of the residuals.
    means :    (K,) array
               Process means. Used as starting values.
    T :        integer
               Desired length of the output time series.
    S :        (K,K) array, optional
               Used as multiplicative change of the disturbance vector to
               increase the variance of the output.
    m :        (K,T) array_like, optional
               Used as additive change during simulation to increase mean of
               the output.
    n_sim_multiple : integer
                    Generate n_sim_multiple * T timesteps. Only the last T
                    timesteps will be returned.
    ia :       (K,T) array_like, optional
               Interannual variability. Used as an additive change during
               simulation to get time-dependent disturbances.
    m_trend :  (K,) array_like, optional
               Used as additive change gradient during simulation to increase
               mean of the output gradually.
    fixed_data : (K,T) array_like, optional
                Keeps the provided time-series fixed. Use np.nans to signify
                values that are not fixed.
                Can be used to simulate hierarchically.

    Returns
    -------
    out :    (K, T) ndarray
             K-dimensional simulated time series.

    """
    K = AM.shape[0]
    if S is None:
        S = np.identity(K, dtype=float)
    n_sim_steps = n_sim_multiple * T + p
    if m is None:
        m = np.zeros((K, n_sim_steps))
    else:
        m = _scale_additive(m, AM, p)
        # we have to expand m to include the pre-simulation timesteps
        m = np.tile(m, n_sim_multiple)
    m_trend = np.asarray([0] * K) if m_trend is None else np.asarray(m_trend)
    m_trend = _scale_additive(m_trend, AM, p)

    if ia is not None:
        ia = _scale_additive(ia, AM, p)

    # the first p columns are initial values, which will be omitted later
    Y = np.zeros((K, n_sim_steps))
    Y[:, :p] = means.reshape((K, -1))

    Y[:, -m.shape[1] :] += m
    start_t = Y.shape[1] - T
    ut = np.array(
        [varwg.rng.multivariate_normal(K * [0], sigma_u) for i in range(q)]
    ).reshape((K, q))
    for t in range(p, n_sim_steps):
        # shift the old values back and draw a new random vector
        ut[:, :-1] = ut[:, 1:]
        ut[:, -1] = varwg.rng.multivariate_normal(K * [0], sigma_u)
        Y[:, t] = ut[:, -1][np.newaxis, :]

        # non-standard scenario stuff
        Y[:, t] = S @ Y[:, t]
        if t > start_t:
            if ia is not None:
                Y[:, t] += ia[:, t - start_t].T
            # apply changes as a trend
            Y[:, t] += float(t - start_t) / T * m_trend

        # conventional VARMA things
        for i in range(p):
            Ai = AM[:, i * K : (i + 1) * K]
            Y[:, t] += Ai @ Y[:, t - i - 1]
        for i in range(p, p + q):
            Mi = AM[:, i * K : (i + 1) * K]
            Y[:, t] += Mi @ ut[:, -1 - i + p]

        if (fixed_data is not None) and (t >= start_t):
            # fixing what's asked to be held constant
            Y[:, t] = np.where(
                np.isnan(fixed_data[:, t - start_t]),
                Y[:, t],
                fixed_data[:, t - start_t],
            )

        if fixed_data is None:
            Y[:, t] += means

    return Y[:, -T:]


def vec(A):
    """The vec operator stacks 2dim matrices into 1dim vectors column-wise.
    See p.661f.

    >>> A = np.arange(6).reshape(2, 3)
    >>> A
    array([[0, 1, 2],
           [3, 4, 5]])
    >>> vec(A)
    array([[0],
           [3],
           [1],
           [4],
           [2],
           [5]])
    """
    return A.T.ravel()[:, None]


def unvec(sequence, K):
    """The inverse of vec.

    >>> a = list(range(6))
    >>> a
    [0, 1, 2, 3, 4, 5]
    >>> unvec(a, K=2)
    array([[0, 2, 4],
           [1, 3, 5]])
    >>> A = np.array([[0, 3, 1, 4, 2, 5]]).T
    >>> unvec(A, K=2)
    array([[0, 1, 2],
           [3, 4, 5]])
    """
    return np.array(sequence).reshape(K, -1, order="F")


def vech(A):
    """The vech operator removes the upper triangular part of a matrix and
    returns the rest in a column-stacked form.
    See p.661f.

    >>> A = np.arange(4).reshape(2, 2)
    >>> A
    array([[0, 1],
           [2, 3]])
    >>> vech(A)
    array([[0],
           [2],
           [3]])
    """
    rows, columns = np.mgrid[0 : A.shape[0], 0 : A.shape[1]]
    return A.T[rows.T >= columns.T][:, None]


def unvech(sequence, K):
    """The inverse of vech.

    >>> a = np.array([0, 2, 3])
    >>> unvech(a, K=2)
    array([[0., 2.],
           [2., 3.]])
    """
    A = np.empty((K, K))
    A[np.tril_indices_from(A)] = np.squeeze(sequence)
    A[np.triu_indices_from(A, k=1)] = A[np.tril_indices_from(A, k=-1)]
    return A


def symmetrize(A):
    return 0.5 * (A + A.T)


def matrix_log_spd(A, eps=1e-10):
    eigvals, eigvecs = np.linalg.eigh(A)
    eigvals_clipped = np.clip(eigvals, eps, None)
    return eigvecs @ np.diag(np.log(eigvals_clipped)) @ eigvecs.T


def matrix_exp_sym(A):
    eigvals, eigvecs = np.linalg.eigh(A)
    return eigvecs @ np.diag(np.exp(eigvals)) @ eigvecs.T


def SC(sigma_u, p, T):
    """Schwarz criterion for VAR_LS order selection (p.150). To be minimized."""
    K = sigma_u.shape[0]
    return np.log(np.linalg.det(sigma_u)) + np.log(T) / T * p * K**2


def HQ(sigma_u, p, T):
    """Hannan-Quinn for VAR_LS order selection (p.150). To be minimized."""
    K = sigma_u.shape[0]
    return np.log(np.linalg.det(sigma_u)) + np.log(np.log(T)) / T * p * K**2


def AIC(sigma_u, p, T):
    """Akaike Information Criterion for order selection of a VAR process.
    See p.147"""
    K = sigma_u.shape[0]
    return np.log(np.linalg.det(sigma_u)) + (2 * p * K**2) / T


def FPE(sigma_u, p, T):
    """Final prediction error.
    See p.147"""
    K = sigma_u.shape[0]
    return ((T + p * K + 1) / (T - p * K - 1)) ** K * np.linalg.det(sigma_u)


def VAR_order_selection(
    data, p_max=10, criterion=SC, estimator=VAR_LS, est_kwds=None
):
    """Order selection for VAR processes to allow parsimonious
    parameterization.

    Parameters
    ----------
    data : (K, T) ndarray
        Input data with K variables and T timesteps.
    p_max : int, optional
        Maximum number of autoregressive order to evaluate.
    criterion : function, optional
        Information criterion that accepts sigma_u, p, and T and returns
        something that gives small values for a parsimonious set of these
        parameters.

    Returns
    -------
    p : int
        Suggested autoregressive order.

    See also
    --------
    AIC : Akaike Information criterion
    FPE : Final Prediction Error
    HQ : Hannan-Quinn information criterion
    SC : Schwartz Criterion
    VAR_LS : Least-squares estimator.
    VAR_residuals : Returns the residuals based on given data and LS estimator
    VAR_LS_sim : Simulation based on LS estimator.
    VAR_LS_predict : Predict given prior data and LS estimator.

    """
    T = data.shape[1]
    if est_kwds is None:
        est_kwds = {}
    return np.argmin(
        [
            criterion(estimator(data, p, **est_kwds)[1], p, T)
            for p in range(p_max + 1)
        ]
    )


def VARMA_order_selection(
    data, p_max=5, q_max=5, criterion=SC, plot_table=False, *args, **kwds
):
    """Returns p and q, the orders of a VARMA process that allows for
    parsimonious parameterization.
    Naive extension of VAR_order_selection without a theoretical basis!"""
    K, T = data.shape
    sigma_us = np.nan * np.empty((p_max + 1, q_max + 1, K, K))
    # we ignore the cases where either p or q is 0, because VARMA_LS_prelim
    # chokes on that
    for p in range(1, p_max + 1):
        for q in range(q_max + 1):
            if q == 0:
                sigma_us[p, q] = VAR_LS(data, p, *args, **kwds)[1]
            sigma_us[p, q] = VARMA_LS_prelim(data, p, q, *args, **kwds)[1]

    crits = list(criterion)
    criterion_table = np.nan * np.empty((len(crits), p_max + 1, q_max + 1))
    for crit_i, crit in enumerate(crits):
        for p in range(1, p_max + 1):
            for q in range(q_max + 1):
                criterion_table[crit_i, p, q] = crit(sigma_us[p, q], p + q, T)

    if plot_table:
        for crit_i, crit in enumerate(crits):
            ts.matr_img(
                criterion_table[crit_i],
                "Information criterion table. %s" % repr(crit),
            )
            ts.plt.xlabel("q")
            ts.plt.ylabel("p")

    p_mins, q_mins = list(
        zip(
            *[
                np.unravel_index(
                    np.nanargmin(criterion_table[ii]),
                    criterion_table[ii].shape,
                )
                for ii in range(len(crits))
            ]
        )
    )
    return p_mins, q_mins, criterion_table


def _scale_additive(additive, A, p=None):
    """Scale an additive online component of a simulation. This prevents
    the overshooting due to auto- and crosscorrelations.

    Parameters
    ----------
    additive :   (K,) or (K,T) ndarray
                 Additive component to be scaled. K is the number of variables
                 simulated.Can be m, m_trend or ia of VARMA_LS_sim, for
                 example.
    A :          (K,K*p) ndarray
                 Parameters of the VAR process. p is the order of the VAR
                 process.
    p :          int, optional
                 Order of the VAR process. If given, only the first K*p columns
                 of A will be interpreted as the parameters of the VAR process.
                 Allows AM to be given as A, which also includes the VMA
                 parameters.

    Returns
    -------
    additive :   (K,) or (K,T) ndarray
                 Scaled additive component.
    """

    A = np.asarray(A)
    K = A.shape[0]
    if p is None:
        p = int(A.shape[1] / K)
        if p * K != A.shape[1]:
            raise ValueError("A is not (K,K*p)-shape.")

    scale_matrix = np.identity(K)
    for i in range(p):
        scale_matrix -= A[:, i * K : (i + 1) * K]
    return scale_matrix @ additive


###############################################################################
## WARNING! The following functions were NOT tested thoroughly!!!!!!!!!!!!!!!!!
###############################################################################


def VAR_LS_predict(data_past, B, sigma_u, T=1, n_realizations=1):
    """Based on a least squares estimator, predict a time-series of the form
    ..math::y(t) = nu + A1*y(t-1) + ... + Ap*y(t-p) + ut
    B contains (nu, A1, ..., Ap).

    Parameters
    ----------
    data_past :      (K, p) ndarray
    B :              (K, p * K + 1) ndarray
                     Parameters of the VAR-process of order p.
    sigma_u :        (K, K) ndarray
                     Covariance matrix of the residuals of the VAR-process.
    T :              int
                     Number of timesteps to predict.
    n_realizations : int
                     Number of realizations. If > 1, gaussian disturbances
                     are added. So if n_realizations=1, the prediction is a
                     best guess.

    Returns
    -------
    Y :             (K, T) or (K, T, n_realizations) ndarray

    References
    ----------
    See p. 707f"""
    # number of variables
    K = B.shape[0]
    # order of VAR_LS-process
    p = (B.shape[1] - 1) / K
    nu = B[:, 0].ravel()

    # the first p columns are initial values, which will be omitted later
    Y = np.zeros((K, data_past.shape[1] + T, n_realizations))
    Y[:, :-T] = data_past[..., np.newaxis]

    for t in range(Y.shape[1] - T, Y.shape[1]):
        for r in range(n_realizations):
            Y[:, t, r] = nu
            if n_realizations > 1:
                Y[:, t, r] += varwg.rng.multivariate_normal(K * [0], sigma_u)

            for i in range(p):
                Ai = B[:, 1 + i * K : 1 + (i + 1) * K]
                Y[:, t, r] += np.squeeze(Ai @ Y[:, t - i - 1, r])

    return np.squeeze(Y[:, -T:])


def VAR_YW(data, p=2):
    """Yule-Walker parameter estimation for a vector auto-regressive model of
    the form Y^0 = A*X + U
    Refer to p. 83ff.
    Here we assume that the data is already mean-adjusted!
    """
    # number of variables
    K, T = data.shape[0], data.shape[1] - p
    # Y is a (K, T) array
    Y = data[:, p:]

    # X is nearly the same as Z in VAR_LS, but without the first row of ones
    X = np.empty((K * p, T))
    Xt = np.empty((K * p, 1))
    for t in range(p, T + p):
        for subt in range(p):
            # HACK! check the p
            Xt[subt * K : (subt + 1) * K] = data[:, t - subt - 1].reshape(
                (K, 1)
            )
        X[:, t - p] = Xt

    # A contains all the parameters (A1, ..., Ap)
    A = np.empty((K, K * p))
    Gamma_y = np.empty_like(A)
    # cov-matrices for up to p lags
    for lag in range(1, p + 1):
        # unbiased cross-covariance
        cov = ts.cross_cov(data, lag) / (T + p - lag)
        start_i = (lag - 1) * cov.shape[0]
        stop_i = lag * cov.shape[0]
        Gamma_y[:, start_i:stop_i] = cov
    Gamma_Y = np.empty((K * p, K * p))
    for ii in range(p):
        start_i = ii * K
        stop_i = (ii + 1) * K
        for jj in range(p):
            start_j = jj * K
            stop_j = (jj + 1) * K
            cov = ts.cross_cov(data, ii - jj) / (T + p - abs(ii - jj))
            Gamma_Y[start_i:stop_i, start_j:stop_j] = cov
    A = Gamma_y @ Gamma_Y.I

    # lets use the same noise as VAR_LS. no idea if this is justified
    sigma_u = Y @ Y.T - Y @ X.T @ np.linalg.inv(X @ X.T) @ X @ Y.T
    sigma_u /= T - K * p - 1

    #    A_dash = Y * X.T * (X * X.T).I
    #    matr_img(np.asarray(A), "A")
    #    matr_img(np.asarray(A_dash), "A_dash")
    #    plt.show()
    return A, sigma_u


def VAR_YW_sim(A, sigma_u, T):
    """Based on a Yule-Walker estimator, simulate a time-series of the form
    ..math:: y(t) = A_1*y(t-1) + ... + A_p*y(t-p) + u(t)
    A contains (A1, ..., Ap).
    See p. 707f"""
    # number of variables
    K = A.shape[0]
    # order of VAR-process
    p = A.shape[1] / K

    # the first p columns are initial values, which will be omitted later
    Y = np.zeros((K, T + p))

    for t in range(p, T + p):
        ut = varwg.rng.multivariate_normal(K * [0], sigma_u).reshape(K, 1)
        Y[:, t] = ut
        for i in range(p):
            Ai = A[:, i * K : (i + 1) * K]
            Y[:, t] += Ai @ Y[:, t - i]

    return Y[:, p:]


# def VAR_YW_residuals(data, A, p=2):
#    K, T = data.shape[0], data.shape[1] - p
#    resi = np.copy(data)
#    for t in xrange(p, T + p):
#        for i in range(p):
#            Ai = A[:, i * K: (i + 1) * K]
#            resi[:, t] -= \
#                np.squeeze(np.asarray(Ai * data[:, t - i - 1].reshape(K, 1)))
#    return resi


def _ut_gamma_part(data, p, q, AM, ut):
    """Recursive calculation of the partial derivatives del ut /del gamma.
    See Lemma 12.1 p.468"""
    K, T = data.shape
    N = K**2 * (p + q)

    Y = data[:, p:]
    R = np.identity(N)
    A_0 = np.identity(K)
    IK_zeros = np.zeros((K**2, N))
    IK_zeros[: K**2, : K**2] = np.identity(K**2)
    zero_IK = np.identity(N)
    ut_gamma_part = np.zeros((K, N, T + p))

    for t in range(p, T - p):
        varma = np.zeros((K, 1))
        for i in range(p):
            Ai = AM[:, i * K : (i + 1) * K]
            varma += Ai @ Y[:, t - i]
        for i in range(p, p + q):
            Mi = AM[:, i * K : (i + 1) * K]
            varma += Mi @ ut[:, -1 - i + p, np.newaxis]

        prev_yu = np.empty(K * (p + q))
        for i in range(p):
            prev_yu[i * K : (i + 1) * K] = Y[:, t - i].T
        for i in range(p, p + q):
            prev_yu[i * K : (i + 1) * K] = ut[:, t - i].T
        M_gamma_part = np.zeros((K, N))
        for i in range(p, p + q):
            Mi = AM[:, i * K : (i + 1) * K]
            M_gamma_part += Mi @ ut_gamma_part[..., t - i + p]
        ut_gamma_part[..., t] = (
            (A_0 @ kron(varma.T, A_0.T)) @ IK_zeros @ R
            - kron(prev_yu, A_0.I) @ zero_IK @ R
            - A_0 @ M_gamma_part
        )
        # ((A_0.I * kron(varma.T, A_0.T)) * IK_zeros * R -
        #  kron(prev_yu, A_0.I) * zero_IK * R -
        #  A_0.I * M_gamma_part)
    return ut_gamma_part


def VARMA_LS(data, p, q, rel_change=1e-3):
    """Implementation of the scoring algorithm to fit a VARMA model. p.470ff"""
    AM_pre, sigma_u_pre = VARMA_LS_prelim(data, p, q)[:2]
    # do not trust the estimator of the residuals
    ut = VARMA_residuals(data, AM_pre, p, q)
    K, T = data.shape
    N = K**2 * (p + q)

    det_new = np.linalg.det(sigma_u_pre)
    # set det_old to something that will cause the while loop to execute at
    # least one time
    det_old = rel_change**-1 * det_new
    print(
        "Determinant of preliminary residual covariance matrix: %f" % det_old
    )

    AM = AM_pre
    gamma = vec(AM)

    ts.matr_img(AM, "AM p=%d q=%d Preliminary" % (p, q))
    ii = 0
    while (det_new > 1e-60) and (
        np.abs(det_old - det_new) / det_old > rel_change
    ):
        ut_gamma_part = _ut_gamma_part(data, p, q, AM, ut)
        sigma_u_gamma = T**-1 * np.sum(
            [ut[:, t] * ut[:, t].T for t in range(ut.shape[1])],  # np.newaxis]
            axis=0,
        )
        sigma_u_gamma_inv = np.linalg.inv(sigma_u_gamma)
        # information matrix
        IM = np.sum(
            [
                ut_gamma_part[..., t].T
                @ sigma_u_gamma_inv
                @ ut_gamma_part[..., t]
                for t in range(T)
            ],
            axis=0,
        )
        likeli_gamma_part = np.sum(
            [
                ut[:, t, np.newaxis].T
                @ sigma_u_gamma_inv
                @ ut_gamma_part[..., t]
                for t in range(ut.shape[1])
            ],
            axis=0,
        )
        gamma -= np.linalg.inv(IM) @ likeli_gamma_part.T

        AM = gamma.reshape((K, K * (p + q)), order="F")
        ut = VARMA_residuals(data, AM, p, q)
        Y = data  # [:, p:])
        X = np.ones((K * (p + q), T))
        Xt = np.zeros((K * (p + q), 1))
        for t in range(p, T):
            for subt in range(p):
                start_i = subt * K
                stop_i = (subt + 1) * K
                Xt[start_i:stop_i] = data[:, t - subt - 1].reshape((K, 1))
            for subt in range(p, p + q):
                start_i = subt * K
                stop_i = (subt + 1) * K
                Xt[start_i:stop_i] = ut[:, t - subt - p - 1].reshape((K, 1))
            X[:, t - p] = Xt
        IK = np.identity(K)
        R = np.identity(N)
        gamma2 = (
            np.linalg.inv(R.T @ kron(X @ X.T, IK) @ R)
            @ R.T
            @ kron(X, IK)
            @ vec(Y)
        )
        residuals_arma_vec = vec(Y) - kron(X.T, IK) @ R @ gamma2
        residuals_arma = residuals_arma_vec.reshape((K, T), order="F")
        # make the residuals have the same length as the data
        ut = np.concatenate((np.zeros((K, p)), residuals_arma), axis=1)
        sigma_u = residuals_arma @ residuals_arma.T / T

        det_new, det_old = np.linalg.det(sigma_u), det_new
        print("Determinant of residual covariance matrix: %f" % det_new)
        ii += 1
        if ii > 1:
            ts.matr_img(
                np.asarray(gamma2.reshape((K, K * (p + q)), order="F")),
                "AM p=%d q=%d Iteration: %d" % (p, q, ii),
            )

    AM = gamma2.reshape((K, K * (p + q)), order="F")
    return AM, sigma_u, ut


def VARMA_residuals(data, AM, p, q):
    K, T = data.shape[0], data.shape[1] - p
    resi = np.copy(data)
    for t in range(p, T + p):
        for i in range(p):
            Ai = AM[:, i * K : (i + 1) * K]
            resi[:, t] -= Ai @ data[:, t - i - 1].reshape(K, -1)
        for i in range(p, p + q):
            Mi = AM[:, i * K : (i + 1) * K]
            resi[:, t] -= Mi @ resi[:, t - i + p - 1].reshape(K, -1)
    return resi


if __name__ == "__main__":
    import doctest

    doctest.testmod()

    import tempfile
    import os
    import varwg

    varwg.conf = varwg.base.conf = varwg.config_template
    p = 2
    T = 2 * 365
    var_names = (
        # we do not use precipitation here as long as we cannot
        # disaggregate it properly
        # "R",
        "theta",
        "Qsw",
        "ILWR",
        "rh",
        # "u", "v"
    )
    met_vg = varwg.VG(var_names)
    met_vg.fit(p, extro=True)



"""
Meteorological Conversions (:mod:`meteo.meteox2y`)
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

The meteox2y module provides functions to calculate derived meteorological
variables from measured meteorological values.

.. currentmodule:: varwg.meteo.meteox2y

.. autosummary::
   :nosignatures:
   :toctree: generated/

   sat_vap_p
   rel2vap_p
   vap_p2rel
   dewpoint
   dew2rel
   norm_pressure
   iziomon
   lw2clouds
   lw_tennessee
   haude
   turc
   turc_rad
   hargreaves
   penman_monteith
   pot_s_rad
   sunshine
   blackbody_rad
   altitude
   spec_hum
   psychro2e
   slope_sat_p

"""

import warnings
from datetime import datetime, timedelta

import bottleneck
import numpy as np
import pytz
import xarray as xr
from scipy import stats
from timezonefinder import TimezoneFinder

import varwg
from varwg import helpers as my

tzf = TimezoneFinder()


def sat_vap_p(at):
    """saturation vapour pressure from air temperature

    Parameters
    ----------
    at : float or numpy.array of floats
        air temperature [deg C]

    Returns
    -------
    c_e : float or numpy.array of floats
        saturation vapour pressure [hPa]

    References
    ----------
    Hydrologie-Skript I, p. 17

    Examples
    --------
    >>> sat_vap_p(25.0)
    31.688149728170984

    >>> at = np.array((0.0, 2.5, 5.0, 7.5, 10.0))
    >>> sat_vap_p(at[:3])
    array([6.11      ,  7.31533365,  8.72596589])
    >>> sat_vap_p(at[3:])
    array([10.3711941 ,  12.28364703])
    """
    at = np.array(at)
    e0 = 6.11
    a = 17.27
    b = 237.3
    c_e = e0 * np.exp(a * at / (b + at))
    return c_e


def rel2vap_p(rh, at):
    """vapour pressure from relative humidity and air temperature

    Parameters
    ----------
    rh : float or numpy.array of floats
        relative humidity with values between 0 and 1 [-]
    at : float or numpy.array of floats
        air temperature [deg C]

    Returns
    -------
    e : float or numpy.array of floats
        vapour pressure [hPa]

    Examples
    --------
    >>> rel2vap_p(0.5,20.0)
    11.695234581996313

    >>> rh = np.array((0.9, 0.8, 0.7, 0.6, 0.5))
    >>> at = np.array((0.0, 2.5, 5.0, 7.5, 10.0))
    >>> rel2vap_p(rh,at)
    array([5.499     ,  5.85226692,  6.10817613,  6.22271646,  6.14182351])

    """
    rh, at = np.array(rh), np.array(at)
    c_e = sat_vap_p(at)
    e = rh * c_e
    return e


def rel2abs_hum(rh, at):
    """Absolute humidity from relative humidity and air temperature.

    Parameters
    ----------
    rh : float or numpy.array of floats
        relative humidity with values between 0 and 1
    at : float or numpy.array of floats
        air temperature [deg C]

    Returns
    -------
    abs_hum : float or numpy.array of floats
        absolute humidity [g / m^3]

    Examples
    --------
    >>> rel2abs_hum(0.8, 20)
    13.831072059995932
    """
    rh, at = map(np.array, (rh, at))
    e = rel2vap_p(rh, at) * 100
    return 2.16679 * e / (273.15 + at)


def abs_hum2rel(abs_hum, at):
    """Relative humidity from absolute humidity and air temperature.

    Parameters
    ----------
    abs_hum : float or numpy.array of floats
        absolute humidity [g / m^3]
    at : float or numpy.array of floats
        air temperature [deg C]

    Returns
    -------
    rh : float or numpy.array of floats
        relative humidity with values between 0 and 1

    Examples
    --------
    >>> abs_hum2rel(13.831072059995932, 20)
    0.8
    """
    abs_hum, at = map(np.array, (abs_hum, at))
    e = abs_hum / 2.16679 * (273.15 + at)
    return vap_p2rel(e / 100, at)


def vap_p2rel(e, at):
    """relative humidity from vapour pressure and air temperature

    Parameters
    ----------
    e : float or numpy.array of floats
        vapour pressure [hPa]
    at : float or numpy.array of floats
        air temperature [deg C]

    Returns
    -------
    rh : float or numpy.array of floats
        relative humidity with values between 0 and 1 [-]

    Examples
    --------
    >>> vap_p2rel(6.22,20.0)
    0.26592027532201407

    >>> e = np.array((5.5, 5.8, 6.1, 6.2, 6.3))
    >>> at = np.array((0.0, 2.5, 5.0, 7.5, 10.0))
    >>> vap_p2rel(e,at)
    array([0.90016367, 0.79285516, 0.69906301, 0.59780966, 0.512877  ])
    """
    e, at = np.array(e), np.array(at)
    c_e = sat_vap_p(at)
    rh = e / c_e
    return rh


def dewpoint(at, rh=None, e=None):
    """dewpoint from air temperature and humidity

    As input is required: air temperature (`at`) and EITHER relative humidity
    (`rh`) OR vapour pressure (`e`).

    Parameters
    ----------
    at : float or numpy.array of floats
        air temperature [deg C]
    rh : float or numpy.array of floats or None
        relative humidity with values between 0 and 1 [-]
    e : float or numpy.array of floats or None
        vapour pressure [hPa]

    Returns
    -------
    dew : float or numpy.array of floats
        dewpoint [deg C]

    Raises
    ------
    Warning
        If relative humidity is > 1.1 (e. g. if vapour pressure is taken as
        relative humidity)

    Examples
    --------
    >>> dewpoint(20.,rh=0.5)
    9.269628637124908

    >>> dewpoint(20.,e=10.0)
    6.968196840688138

    >>> at = np.array((0.0, 2.5, 5.0, 7.5, 10.0))
    >>> rh = np.array((0.9, 0.8, 0.7, 0.6, 0.5))
    >>> dewpoint(at, rh=rh)
    array([-1.43893707, -0.59071344, -0.0041022 ,  0.25144096,  0.07140267])

    >>> at = np.array((0.0, 2.5, 5.0, 7.5, 10.0))
    >>> e = np.array((5.5, 5.8, 6.1, 6.2, 6.3))
    >>> dewpoint(at, e=e)
    array([-1.43646874, -0.71330622, -0.02250498,  0.20109231,  0.42152362])
    """
    at = np.array(at)
    e0 = 6.11
    a = 17.27
    b = 237.3
    if rh is not None:
        e = rel2vap_p(rh, at)
        summ = np.sum(rh > 1.1)
        if summ > 0:
            warnings.warn(" %i relative humidity values are > 1.1" % summ)
    dew = b * (np.log(e) - np.log(e0)) / (a - (np.log(e) - np.log(e0)))
    return dew


def dew2rel(dew, at):
    """relative humidity from dewpoint and air temperature

    Parameters
    ----------
    at : float or numpy.array of floats
        air temperature [deg C]
    dew : float or numpy.array of floats
        dewpoint [deg C]

    Returns
    -------
    rh : float or numpy.array of floats
        relative humidity with values between 0 and 1 [-]

    Examples
    --------
    >>> dew2rel(15.,16.0)
    0.9378863357566809

    >>> at = np.array((0.0, 2.5, 5.0, 7.5, 10.0))
    >>> dew = np.array((0.0, 3., 5.0, 7.7, 10.0))
    >>> dew2rel(at,dew)
    array([1.        ,  0.96506519,  1.        ,  0.98642692,  1.        ])
    """
    dew, at = np.array(dew), np.array(at)
    e = sat_vap_p(dew)
    c_e = sat_vap_p(at)
    rh = e / c_e
    return rh


def norm_pressure(p, at, h=454.0):
    """normalize pressure to sealevel

    formula from wikipedia.de for linear temperature gradient 0.0065K/m

    Parameters
    ----------
    p : float or numpy.array of floats
        pressure [hPa]
    at : float or numpy.array of floats
        air temperature [deg C]
    h : float, optional
        height above sea level of measuring station [m]
        default value = 454.0 (height of station Stuttgart-Lauchaecker)

    Returns
    -------
    p_nn : float or numpy.array of floats
        sea level pressure [hPa]

    References
    ----------

    Examples
    --------
    >>> "%.9f" % norm_pressure(930,10.0)
    '982.074383073'

    >>> "%.9f" % norm_pressure(930,10.0,h=765)
    '1019.090035371'

    >>> at = np.array((0.0, 2.5, 5.0, 7.5, 10.0))
    >>> p = np.array((925, 930, 935, 932, 931))
    >>> norm_pressure(p,at,h=765)
    array([1016.98077228,  1021.60706988,  1026.24032953,  1022.10690872,
            1020.18583111])
    """
    tk = at + 273.16  # T in Kelvin
    p_nn = p * (1 + 0.0065 * h / tk) ** 5.255
    return p_nn


def iziomon(temp, clouds, rh=None, dew=None, e=None, site="low"):
    """incident long wave radiation following Iziomon et al (2003)[1]_

    As input is required: air temperature (`temp`), cloud cover (`clouds`) and
    EITHER relative humidity (`rh`) OR dewpoint (`dew`) OR vapour pressure
    (`e`).

    Parameters
    ----------
    temp : float or numpy.array of floats
        air temperature [deg C]
    clouds : float or numpy.array of floats
        cloud cover with values between 0 and 1 [-]
    rh : float or numpy.array of floats or None
        relative humidity with values between 0 and 1 [-]
    dew : float or numpy.array of floats or None
        dewpoint [deg C]
    e : float or numpy.array of floats or None
        vapour pressure [hPa]
    site : {'low', 'high'}
        parameterisation for lowland or highland site

    Returns
    -------
    lw : float or numpy.array of floats
        incident longwave radiation [W/m**2]

    See Also
    --------
    lw2clouds : reverse (get cloud cover out of long wave, temperature and
        humidity)

    Notes
    -----
    Empirical formula, found for experiments in Bremgarten (47deg54'35''N;
    7deg37'18''E) in the Upper Rhine plain in Germany (lowland site) and
    Feldberg, 1489 m asl, 47deg52'31''N, 8deg00'11''E, Black Forest, Germany
    (highland site)

    References
    ----------
    .. [1] Iziomon, M.G., Mayer H, Matzarakis A. (2003): Downward atmospheric
       longwave irradiance under clear and cloudy skies: Measurement and
       parameterization, Journal of Atmospheric and Solar-Terrestrial Physics
       65 (2003) 1107 - 1116

    Examples
    --------
    >>> iziomon(15.,0.5,rh=0.89)
    327.52426791875763

    >>> temp = np.array((0.0, 2.5, 5.0, 7.5, 10.0))
    >>> clouds = np.array((0.0, 0.1, 0.8, 0.5, 1.0))
    >>> e = np.array((5.5, 5.8, 6.1, 6.2, 6.3))
    >>> iziomon(temp,clouds,e=e)
    array([225.34414848,  235.0777488 ,  279.01405432,  267.25348612,
            321.15448959])
    """
    # iziomon-parameter lowland site:
    if site == "low":
        Xs, Ys, Zs = 0.35, 10.0, 0.0035
    # mountain site:
    elif site == "high":
        Xs, Ys, Zs = 0.43, 11.5, 0.005
    # temp in Kelvin:
    tk = temp + 273.16
    # humidity in vapour pressure:
    if rh is not None:
        e = rel2vap_p(rh, temp)
    elif dew is not None:
        e = sat_vap_p(dew)
    # clouds in Okta:
    clouds = clouds * 8
    # iziomon:
    lw_clear = blackbody_rad(temp=temp) * (1 - Xs * np.exp(-Ys * e / tk))
    lw = lw_clear * (1 + Zs * clouds**2)
    return lw


def temp2lw(temp):
    """Incident long-wave radiation from air temperature (Gal pc).

    Parameters
    ----------
    temp : float or numpy.array of floats
        air temperature [deg C]

    Returns
    -------
    lw : float or numpy.array of floats
        incident longwave radiation [W/m**2]
    """
    from scipy import constants

    try:
        theta_kelvin = constants.C2K(temp)
    except AttributeError:
        theta_kelvin = constants.convert_temperature(temp, "Celsius", "Kelvin")
    e = 6.2 * np.exp(17.26 * temp / (theta_kelvin - 35.8))
    boltz = 0.0000000567  # ask gideon
    alpha, beta = 0.42, 0.065
    return boltz * theta_kelvin**4 * (alpha + beta * np.sqrt(e))


# dirk: using np.atleast_1d and as scalar here, to make indexing on scalar
# input and the doctests work
@my.asscalar
def lw2clouds(lw, temp, rh=None, dew=None, e=None, site="low"):
    """Cloud cover from incident long wave radiation (Iziomon et al (2003)[1]_)

    As input is required: incident long wave radiation (`lw`),
    air temperature (`temp`) and EITHER relative humidity (`rh`)
    OR dewpoint (`dew`) OR vapour pressure (`e`)

    Parameters
    ----------
    lw : float or numpy.array of floats
        incident longwave radiation [W/m**2]
    temp : float or numpy.array of floats
        air temperature [deg C]
    rh : float or numpy.array of floats or None
        relative humidity with values between 0 and 1 [-]
    dew : float or numpy.array of floats or None
        dewpoint [deg C]
    e : float or numpy.array of floats or None
        vapour pressure [hPa]
    site : {'low', 'high'}
        parameterisation for lowland or highland site

    Returns
    -------
    clouds : float or numpy.array of floats
        cloud cover with values between 0 and 1 [-]

    See Also
    --------
    iziomon : incident long wave radiation from air temperature, cloud cover
        and humidity

    Notes
    -----
    Resulting cloud cover is always between 0 and 1. Gives 0 for unrealistic
    low and 1 for unrealistic high values of `lw`.

    References
    ----------
    .. [1] Iziomon, M.G., Mayer H, Matzarakis A. (2003): Downward atmospheric
       longwave irradiance under clear and cloudy skies: Measurement and
       parameterization, Journal of Atmospheric and Solar-Terrestrial Physics
       65 (2003) 1107 - 1116

    Examples
    --------
    >>> lw2clouds(328., 15., rh=0.9)
    0.49959967427213553

    >>> lw = np.array((225.5, 235.1, 279, 267, 321.2))
    >>> temp = np.array((0.0, 2.5, 5.0, 7.5, 10.0))
    >>> e = np.array((5.5, 5.8, 6.1, 6.2, 6.5))
    >>> lw2clouds(lw,temp,e=e)
    array([0.0555659 ,  0.1020956 ,  0.79983929,  0.49550839,  0.99289638])
    """
    lw, temp = np.atleast_1d(lw, temp)
    # iziomon-parameter lowland site:
    if site == "low":
        Xs, Ys, Zs = 0.35, 10.0, 0.0035
    # mountain site:
    elif site == "high":
        Xs, Ys, Zs = 0.43, 11.5, 0.005
    # temp in Kelvin:
    tk = temp + 273.16
    # humidity in vapour pressure:
    if rh is not None:
        e = np.atleast_1d(rel2vap_p(rh, temp))
    elif dew is not None:
        e = np.atleast_1d(sat_vap_p(dew))
    elif e is not None:
        e = np.atleast_1d(e)
    # iziomon:
    lw_clear = blackbody_rad(temp=temp) * (1 - Xs * np.exp(-Ys * e / tk))
    lw_valid = lw > lw_clear
    n = np.zeros_like(lw)
    n[lw_valid] = ((lw[lw_valid] / lw_clear[lw_valid] - 1) / Zs) ** 0.5
    clouds = n / 8.0
    clouds[np.isnan(clouds)] = 0
    clouds[clouds > 1] = 1
    return clouds


def lw_tennessee(temp, clouds):
    """incident long wave radiation

    Parameters
    ----------
    temp : float or numpy.array of floats
        air temperature [deg C]
    clouds : float or numpy.array of floats
        cloud cover with values between 0 and 1 [-]

    Returns
    -------
    lw : float or numpy.array of floats
        incident longwave radiation [W/m**2]

    See Also
    --------
    iziomon : other empirical formula, from southwestern germany

    References
    ----------
    .. [1] Tennessee Valley Authority 1972. Heat and mass transfer between a
       water surface and the atmosphere Water Resources Research Laboratory
       Report 14, Report No. 0-6803. (I didn't find it. But it's mentioned in:)
    .. [2] ELCOM Science manual
    """
    # temp in Kelvin:
    tk = temp + 273.16
    c_e = 9.37 * 10**-6  # K**-2
    eps_a = c_e * tk**2
    lw = (1 + 0.17 * clouds**2) * eps_a * blackbody_rad(temp=temp)
    return lw


def haude(svp, vp, mon):
    """calculates the evapotranspiration following Haude (1955) [1]_
    as described in the Hydrologie-I-Skript

    Parameters
    ----------
    svp : float or numpy.array of floats
        saturation vapour pressure [hPa], measured at 2pm
    vp : float or numpy.array of floats
        vapour pressure [hPa], measured at 2pm
    mon : int or numpy.array of ints
        month: 1=january, 12=december

    Returns
    -------
    etp : float or numpy.array of floats
        evapotranspiration [mm] daily values

    See Also
    --------
    turc : potential evapotranspiration following Turc

    Notes
    -----
    The Haude formula is only valid in temperate humid climate

    References
    ----------
    .. [1] Haude, W. (1955): Zur Bestimmung der Verdunstung auf moeglichst
       einfache Weise. - Mitt. Dt. Wetterd. 2 (11), Bad Kissingen (Dt.
       Wetterd.)

    Examples
    --------
    >>> haude(12.28,8.83,5)
    1.0005

    >>> haude(12.28,8.83,12)
    0.759

    >>> svp = np.array([11.87809345,  12.28364703,  12.70132647])
    >>> vp = np.array([8.55222728,  8.84422586,  9.14495506])
    >>> haude(svp,vp,12)
    array([0.73169056,  0.75667266,  0.78240171])

    >>> mon = np.array([2,3,4])
    >>> haude(svp,vp,mon)
    array([0.73169056,  0.75667266,  1.03134771])
    """
    mon = mon - 1
    hfs = (
        0.0022,
        0.0022,
        0.0022,
        0.0029,
        0.0029,
        0.0028,
        0.0026,
        0.0029,
        0.0023,
        0.0022,
        0.0022,
        0.0022,
    )  # monthly Haude factor
    try:
        hf = hfs[mon]
    except TypeError:
        hf = np.array([hfs[mo] for mo in mon])
    svp, vp = np.array(svp) * 100, np.array(vp) * 100  # hPa -> Pa
    etp = hf * (svp - vp)
    return etp


def turc(at, ts, mon):
    """calculates the potential evapotranspiration following Turc [1]_ as
    described in the Hydrologie-I-Skript (p. 55)

    Parameters
    ----------
    at : float or numpy.array of floats
        daily average air temperature [degC]
    ts : float or numpy.array of floats
        number of sunshine hours per day
    mon : int or numpy.array of ints
        month: 1=january, 12=december

    Returns
    -------
    etp : float or numpy.array of floats
        potential evapotranspiration [mm] daily values

    See Also
    --------
    haude : evapotranspiration following Haude

    Notes
    -----
    The Turc formula is only valid for `at` > 0 deg C

    References
    ----------
    .. [1] TURC??

    Examples
    --------
    >>> "%.9f" % turc(10,0,5)
    '1.104000000'

    >>> "%.9f" % turc(10,0,12)
    '0.408000000'
    """
    mon = mon - 1
    # Konstanten C1, C2 fuer etwa 53 deg N in Mitteleuropa
    # warum auch immer
    cs = (
        (1.09, 0.18),
        (1.4, 0.259),
        (1.86, 0.35),
        (2.36, 0.429),
        (2.76, 0.476),
        (3, 0.489),
        (2.93, 0.484),
        (2.58, 0.448),
        (2.1, 0.39),
        (1.57, 0.21),
        (1.19, 0.21),
        (1.02, 0.158),
    )
    try:
        c1, c2 = cs[mon]
    except TypeError:
        c1 = np.array([cs[mo][0] for mo in mon])
        c2 = np.array([cs[mo][1] for mo in mon])
    etp = (c1 + c2 * ts) * at / (at + 15)
    return etp


def turc_rad(at, rh, G):
    """calculates the potential evapotranspiration following Turc

    with the global radiation instead of the empiric factors according to
    Hydrologie und Wasserwirtschaft by Maniak

    Parameters
    ----------
    at : numpy.array of floats
        mean daily air temperature [deg C]
    rh : numpy.array of floats
        relative humidity [%]
    G  : numpy.array of floats
        global radiation [W/m^2]

    Returns
    -------
    etp : numpy.array of floats
        potential evapotranspiration [mm/d]

    Notes
    -----
    - For etp < 0.1 mm/d the evapotranspiration rate is set to 0.1 mm/d.
      [DVWK 1996]
    - According to the energetic limit of 7 mm/d in Germany, ETPmax is set to
      7 mm/d.
    """

    rh = np.array(rh, float)
    C = np.ones_like(rh)
    ii = np.where(rh < 50.0)
    C[ii] = 1.0 + ((50.0 - rh[ii]) / 70.0)

    G = G * 86400.0 / 10000.0  # W/m^2 in J/(cm^2*d)

    etp = 0.0031 * C * (G + 209.0) * (at / (at + 15.0))
    ii = np.where(etp > 7.0)
    etp[ii] = 7.0
    jj = np.where(etp < 0.1)
    etp[jj] = 0.1

    return etp


def hargreaves(tmax, tmin, date, in_format="%Y-%m-%dT%H:%M:%S"):
    """calculates the potential evapotranspiration rate [mm/d] following
    Hargreaves & Samani 1985 according to THE ASCE STANDARDIZED REFERENCE
    EVAPOTRANSPIRATION EQUATION

    Parameters
    ----------
    tmax : np.array of floats
        maximum of the daily air temperature [deg C]
    tmin : np.array of floats
        minimum of the daily air temperature [deg C]
    date : np.array of strings
        date strings in format in_format
    in_format : format string
        default: '%Y-%m-%dT%H:%M:%S'

    Returns
    -------
    etp : numpy.array of floats
        potential evapotranspiration [mm/d]

    Notes
    -----
    - equation needs the extraterrestrial radiation Ra, which depends on:

      - inverse relative distance factor for the earth-sun dr []
      - solar declination delta [rad]
      - sunset hour angle [rad]
        with latitude Lauchaecker lat=48.738 deg => pi/180 * 48.738
        deg = 0.8506 rad

    - conversion of date in day of year (j)
    - ETPmax = 7.0 mm/d due to energetic limit (Germany)

    References
    ----------
    """

    j = []
    for dat in date:
        t = datetime.strptime(dat, in_format)
        doy = float(datetime.strftime(t, format="%j"))  # day of year
        j.append(doy)

    # inverse relative distance factor for the earth-sun []
    dr = 1.0 + 0.033 * np.cos(np.multiply((2 * np.pi / 365.0), j))
    # solar declination [rad]
    delta = 0.409 * np.sin(np.multiply((2 * np.pi / 365.0), j) - 1.39)
    # sunset hour angle [rad]
    omega_s = np.arccos(-np.tan(0.8506) * np.tan(delta))
    # extraterrestrial radiation [MJ m^-2 d^-1]
    Ra = (
        24.0
        / np.pi
        * 4.92
        * dr
        * (
            omega_s * np.sin(0.8506) * np.sin(delta)
            + np.cos(0.8506) * np.cos(delta) * np.sin(omega_s)
        )
    )
    # [MJ m^-2 d^-1]
    etp = (
        0.0023
        * (tmax - tmin) ** (0.5)
        * ((tmax + tmin) / 2.0 + 17.8)
        * Ra
        / 2.45
    )  # 2.45: factor for calculating mm/d

    ii = np.where(etp > 7.0)
    etp[ii] = 7.0

    return etp


def penman_monteith(at, u, Rn, rh):
    """Calculates the reference crop evaporation following the Penman-Monteith
    method and the FAO-56 determinations published in ASCE
    Standardized Reference Evapotranspiration Equation

    Parameters
    ----------
    at : np.array of floats
        mean daily air temperature at 2m-height [deg C]
    u  : np.array of floats
        mean daily wind speed at 2m-height [m/s]
    Rn : np.array of floats
        measured net radiation at the crop surface [MJ m^-2 d^-1]
    rh : np.array of floats
        relative humidity [%]

    Returns
    -------
    eto : np.array of floats
        FAO Penman-Monteith standardized reference crop evapotranspiration for
          short (~=0.12m) surfaces [mm d^-1]

    Notes
    -----
    - Units for the 0.408 coefficient are m^2 mm MJ^-1
    - The FAO-56 Penman-Monteith equation is a grass reference equation that
      was derived from the Penman-Monteith form of the combination equation
      (Monteith 1965, 1981) by fixing h = 0.12 m for clipped grass and by
      assuming measurement heights of z = 2 m (at, rh, u) and using a latent
      heat of vaporization of 2.45 MJ kg-1.
      The result is an equation that defines the reference evapotranspiration
      from a hypothetical grass surface having a fixed height of 0.12 m, bulk
      surface resistance of 70 s m-1, and albedo of 0.23.
    - in relationship to the net radiation, the soil heat flux is very small
      and is fixed with G = 0.1*Rn
    """
    rh = rh / 100.0  # relative humidity in decimal[0,1]
    y = 0.000665 * 101.3 * ((293.0 - 0.0065 * 453.0) / 293.0) ** 5.26

    G = 0.1 * Rn
    S = slope_sat_p(at)
    Cn = 900
    Cd = 0.34
    es = sat_vap_p(at)
    ea = rel2vap_p(rh, at)

    eto = (
        0.408 * S * (Rn - G)
        + y * (Cn / (at + 273.15)) * u * ((es - ea) / 10.0)
    ) / (S + y * (1 + Cd * u))

    ii = np.where(eto > 7.0)
    eto[ii] = 7.0

    return eto


def pot_s_rad(
    date,
    lat=48.738,
    longt=9.099,
    in_format="%Y-%m-%dT%H:%M",
    tz_mer=15.0,
    wog=-1,
):
    """theoretical maximal potential solar radiation outside of atmosphere

    Parameters
    ----------
    date : numpy.array of time strings (format: in_format) or datetime objects
        or floats (doys)
    lat : float, optional
        latitude of station in decimal degrees, default: Stuttgart Lauchaecker
    longt : float, optional
        longitude of station in decimal degrees, default: Stuttgart Lauchaecker
    in_format : time format string, optional
        format of date if date is string, default '%Y-%m-%dT%H:%M'
    tz_mer : int, optional
        central meridian of time zone, default: 15 (CET)
    wog : {-1, 1}, optional
        west of greenwich, 1 if west, -1 if east, default = -1

    Returns
    -------
    smax : numpy.array of floats
        maximal potential solar radiation in W/m^2

    Notes
    -----
    from campbell technical note 18 [1]_, except declination of sun (d):
    formula of Spencer (1971) [2]_

    WARNING: Campbell Scientific recommends the use of a high quality sun
    screen lotion when exposing your skin to solar radiation for large values
    of sunshine hours!

    References
    ----------
    .. [1] Campbell Scientific (2005) technical note 18: CALCULATING SUNSHINE
       HOURS FROM PYRANOMETER / SOLARIMETER DATA
    .. [2] Spencer JW (1971) Fourier series representation of the position of
       the Sun. Search 2: 172.

    Examples
    --------
    >>> date_str = np.array(["2011-09-28T11:27"])
    >>> pot_s_rad(date_str)
    array([855.35624182])
    >>> from varwg import times
    >>> dt = times.str2datetime(date_str, "%Y-%m-%dT%H:%M")
    >>> pot_s_rad(dt)
    array([855.35624182])
    >>> pot_s_rad(times.datetime2doy(dt))
    array([855.35624182])
    """
    if tz_mer is None:
        timezone_str = tzf.timezone_at(lat=float(lat), lng=float(longt))
        timezone = pytz.timezone(timezone_str)
        dt = datetime(2020, 6, 1)
        tz_offset = (
            timezone.utcoffset(dt, is_dst=False).total_seconds()
        ) / 3600
        tz_mer = 15 * tz_offset
    s0 = 1373  # Solarkonstante W/m^2
    # once upon a time there was a latitude
    Lc = wog * (tz_mer - longt) / 15.0  # Local correction of time
    lat = lat * np.pi / 180  # in rad
    try:
        # where we used to raise an exception
        doys = varwg.times.datetime2doy(date)  # if date is datetime
    except (
        TypeError,
        AttributeError,
        ValueError,
        NotImplementedError,
        IndexError,
    ):
        try:
            # ...or two
            # if date is string
            doys = varwg.times.datetime2doy(
                varwg.times.str2datetime(date, in_format)
            )
        except (TypeError, IndexError):
            # those were the doys my friend, i thought they never end
            doys = date  # if date is already in doys
    hours = (doys - doys.astype(int)) * 24
    # remember how we typecasted away the hours
    doys = doys.astype(int)
    j1 = doys / 100.0
    j2 = (doys - 180) / 100.0
    Et = np.where(
        doys > 180,
        (
            -0.05039
            - 0.33954 * j2
            + 0.04084 * j2**2
            + 1.8928 * j2**3
            - 1.7619 * j2**4
            + 0.4224 * j2**5
        ),
        (
            -0.04056
            - 0.74503 * j1
            + 0.08823 * j1**2
            + 2.0516 * j1**3
            - 1.8111 * j1**4
            + 0.42832 * j1**5
        ),
    )
    t0 = 12 - Lc - Et
    # we sang and dance forever and a doy (less)
    # la la la la lala lala lala lala
    la = 2 * np.pi / 365 * (doys - 1)
    d = (
        0.006918
        - 0.399912 * np.cos(la)
        + 0.070257 * np.sin(la)
        - 0.006758 * np.cos(2 * la)
        + 0.000907 * np.sin(2 * la)
        - 0.002697 * np.cos(3 * la)
        + 0.00148 * np.sin(3 * la)
    )
    sind = np.sin(d)
    sinphi = np.atleast_1d(
        sind * np.sin(lat)
        + np.cos(d) * np.cos(lat) * np.cos(15 * np.pi / 180.0 * (hours - t0))
    )
    sinphi[sinphi < 0] = 0
    smax = s0 * sinphi
    return smax


def pot_s_rad_daily(
    date, lat=48.738, longt=9.099, in_format="%Y-%m-%d", tz_mer=15.0, wog=-1
):
    """daily average values of --> pot_s_rad"""
    # machen wir mal stundenweise:
    if not isinstance(date[0], datetime):
        date = varwg.times.str2datetime(date, in_format)
    date_h = np.array(
        [dt__ + timedelta(hours=i) for dt__ in date for i in range(24)]
    )
    pot_h = pot_s_rad(date_h, lat, longt, tz_mer=tz_mer, wog=wog)
    return np.average(pot_h.reshape(-1, 24), axis=1)


def sunshine_pot(doys, lat=48.738, longt=9.099, tz_mer=15.0, wog=-1):
    """Maximum daily sunshine hours based on evaluating pot_s_rad per minute."""
    mins_per_day = 24 * 60
    # doys = doys.astype(float)
    doys = doys[doys == doys.astype(int)]
    doys_minutes = doys.repeat(mins_per_day).reshape(-1, mins_per_day)
    doys_minutes += (np.arange(mins_per_day) / float(mins_per_day))[None, :]
    doys_minutes = doys_minutes.ravel()
    smax = pot_s_rad(doys_minutes, lat, longt, tz_mer=tz_mer, wog=wog)
    smax = smax.reshape(-1, mins_per_day)
    sun_hours = np.sum(smax > 0, axis=1) / 60.0
    return sun_hours


def sunshine(
    sw,
    date,
    lat=48.738,
    longt=9.099,
    in_format="%Y-%m-%dT%H:%M",
    tz_mer=15.0,
    wog=-1,
):
    """sunshine or not?

    calculates maximum potential solar radiation depending on latitude,
    longitude, and time and compares it to actual solar radiation. Sunshine if
    actual solar radiation > 0.4*maximum potential solar radiation
    accurate enough for normal non-scientific use of sunshine hour data

    Parameters
    ----------
    sw : numpy.array of floats
        solar (short wave) radiation in W/m^2
    date : numpy.array of time strings
        date and time in format in_format
    lat : float, optional
        latitude of station in decimal degrees, default: Stuttgart Lauchaecker
    longt : float, optional
        longitude of station in decimal degrees, default: Stuttgart Lauchaecker
    in_format : time format string, optional
        format of date, default '%Y-%m-%dT%H:%M'
    tz_mer : int, optional
        central meridian of time zone, default: 15 (CET)
    wog : {-1, 1}, optional
        west of greenwich, 1 if west, -1 if east, default = -1

    Returns
    -------
    shining : numpy.array containing 0 and 1
        1: sun is shining in corresponding time step, 0: sun is not shining in
        corresponding time step

    Examples
    --------
    >>> sunshine(np.array([450]),np.array(["2011-09-28T11:27"]))
    array([1])
    >>> sunshine(np.array([200]),np.array(["2011-09-28T11:27"]))
    array([0])
    """
    shining = np.zeros_like(sw)
    smax = pot_s_rad(
        date, lat, longt, in_format=in_format, tz_mer=tz_mer, wog=wog
    )
    shining[np.where(sw > 0.4 * smax)] = 1
    return shining


def sonnenscheindauer(date, sw, del_t=60):
    r"""
    bestimmt Sonnenscheindauer anhand der kurzwelligen Solarstrahlung

    Parameters
    ----------
    date  : string
        datetime.datetime(2040, 1, 1, 0, 0, 0)
    sw : string
    del_t : int, optional
        timestep in minutes: 1-min-data is averaged to this timestep.
        Default is 60

    Returns
    -------
    sunshine_hour
    """

    sunshine_min = sunshine(sw, date)
    # sunshine_sum = np.sum(sunshine_min)
    sunshine_hour = sunshine_min / float(del_t)

    return sunshine_hour


def get_tz_offset(dates, longitude, latitude):
    # from tzwhere import tzwhere
    # import pytz
    # tzw = tzwhere.tzwhere()
    # timezone offset
    # timezone_str = tzw.tzNameAt(latitude, longitude)
    if isinstance(dates, datetime):
        dates = (dates,)
    timezone_str = tzf.timezone_at(lat=float(latitude), lng=float(longitude))
    timezone = pytz.timezone(timezone_str)
    dts = [datetime(date.year, date.month, date.day) for date in dates]
    tz_offset = (
        np.array(
            [timezone.utcoffset(dt, is_dst=True).total_seconds() for dt in dts]
        )
        / 86400
    )
    return np.squeeze(tz_offset)


def sunshine_hours(dates, longitude, latitude, tz_offset=None):
    """Calculates hours from sunrise to sunset.

    Notes
    -----
    see:
    https://en.wikipedia.org/wiki/Sunrise_equation
    https://michelanders.blogspot.com/2010/12/calulating-sunrise-and-sunset-in-python.html

    Examples
    --------
    >>> from datetime import date
    >>> sunshine_hours([date(2018, 6, 1)], 8.848, 48.943)  # mühlacker
    array([15.89961046])
    """
    if tz_offset is None:
        tz_offset = get_tz_offset(dates, longitude, latitude)
    J_rise, J_set = sunshine_riseset(dates, longitude, latitude, tz_offset)
    # return (J_set - J_rise) * 60
    return J_set - J_rise


def max_sunshine_minutes(dates, longitude, latitude, tz_offset=None):
    if tz_offset is None:
        tz_offset = get_tz_offset(dates, longitude, latitude)
    hour_of_day = np.array([date.hour for date in dates])
    sun_rise, sun_set = sunshine_riseset(dates, longitude, latitude, tz_offset)
    minutes = np.zeros_like(dates, dtype=float)
    minutes[(hour_of_day > sun_rise) & (hour_of_day < sun_set)] = 60
    # if sunrise happens in the hour_of_day, we have to make minutes
    # smaller
    dawn_mask = (hour_of_day < sun_rise) & (sun_rise < hour_of_day + 1)
    minutes[dawn_mask] = 60 * (sun_rise[dawn_mask] - hour_of_day[dawn_mask])
    # equivalent for the evening
    twilight_mask = (hour_of_day < sun_set) & (sun_set < hour_of_day + 1)
    minutes[twilight_mask] = 60 * (
        hour_of_day[twilight_mask] + 1 - sun_set[twilight_mask]
    )
    assert np.all(minutes >= 0)
    assert np.all(minutes <= 60)
    return minutes


def sunshine_riseset(dates, longitude, latitude, tz_offset=None):
    """Calculates sunrise and sunset hours."""
    if tz_offset is None:
        tz_offset = get_tz_offset(dates, longitude, latitude)
    jdn = varwg.times.date2jdn(dates)

    if tz_offset is None:
        tz_offset = get_tz_offset(dates, longitude, latitude)
    n = jdn - 2451545.0008  # current julian day
    J_star = n - longitude / 360 + tz_offset  # mean solar noon
    M = (357.5291 + 0.98560028 * J_star) % 360  # solar mean anomaly
    M_radians = np.radians(M)
    # equation of the center
    C = (
        1.9148 * np.sin(M_radians)
        + 0.02 * np.sin(2 * M_radians)
        + 0.0003 * np.sin(3 * M_radians)
    )
    # ecliptic longitude
    lambda_radians = np.radians((M + C + 282.9372) % 360)
    # solar transit
    J_transit = (
        2451545.5
        + J_star
        + 0.0053 * np.sin(M_radians)
        - 0.0069 * np.sin(2 * lambda_radians)
    )
    # declination of the sun
    sin_del = np.sin(lambda_radians) * np.sin(np.radians(23.44))
    cos_del = np.cos(np.arcsin(sin_del))
    # hour angle
    latitude_radians = np.radians(latitude)
    cos_omega0 = (
        np.sin(np.radians(-0.83)) - np.sin(latitude_radians) * sin_del
    ) / (np.cos(latitude_radians) * cos_del)
    omega0 = np.degrees(np.arccos(cos_omega0)) / 360
    J_transit = J_transit - J_transit.astype(int)
    J_rise = 24 * (J_transit - omega0)
    J_set = 24 * (J_transit + omega0)
    return J_rise, J_set


def blackbody_rad(rad=None, temp=None, eps=1.0):
    """Stefan-Boltzmann law

    There are two ways to use this funtion:

    1) If input is radiation, then the function calculates the absolute
    temperature of the body emitting the radiation

    2) If input is temperature, then the function calculates the radiation
    emitted by the body of this temperature

    Parameters
    ----------
    rad : float or numpy.array of floats or None
        radiation [W/m**2]
    temp : float or numpy.array of floats or None
        temperature [deg C]
    eps : float or numpy.array of floats, optional
        emissivity of a grey body, values between 0 and 1, default=1

    Returns
    -------
    either:
    temp : float or numpy.array of floats or None
        temperature [deg C]
    or:
    rad : float or numpy.array of floats or None
        radiation [W/m**2]

    Examples
    --------
    >>> "%.9f" % blackbody_rad(rad=350)
    '7.138805085'

    >>> "%.9f" % blackbody_rad(temp=0)
    '315.683203500'
    """
    sigma = 5.67 * 10**-8
    tk0 = 273.16
    if rad is not None:
        return (rad / (eps * sigma)) ** 0.25 - tk0
    elif temp is not None:
        return eps * sigma * (temp + tk0) ** 4


def altitude(temp1, temp0, pres1, pres0, alt0):
    r"""Converts continous meassured vertical pressure and temperature data to
    altitude.
    # 1 means i and 0 means i-1

    Parameters
    ----------
    temp1 : float or array_like
        Temperature value at timestep i.
    temp0 : float or array_like
        Temperature value at timestep i-1.
    pres1 : float or array_like
        Pressure value at timestep i.
    pres0 : float or array_like
        Pressure value at timestep i-1.
    alt0 : float or array_like
        Altitude value at timestep i-1.

    Returns
    -------
    altitude : float or array_like
        Altitude at timestep i

    Notes
    -----
    This is the implemented equation:

    g : Gravity acceleration 9.81 m/s^2
    T_0 : Temperature 273.16 K

    .. math::

        altitude = ((\frac{\ln(pres0)}{pres1})*287*((\frac{(\frac{(temp1+temp0)}{2})+T_0)}/{g}))+alt0

    References
    ----------
    *** -> Ask Felix!!! ***

    Examples
    --------
    >>> alt = np.nan*np.ones(40)
    >>> alt[0] = 600
    >>> for i,element in enumerate(alt):    # doctest: +SKIP
    ...    if str(element) == 'nan':    # doctest: +SKIP
    ...        element = altitude(temp[i], temp[i-1], pres[i], pres[i-1], alt[i-1])    # doctest: +SKIP
    """
    tk0 = 273.16
    g = 9.81

    altitude = (
        (np.log(pres0 / pres1)) * 287 * ((((temp0 + temp1) / 2) + tk0) / g)
    ) + alt0
    # ;print 'altitude: ',self.altitude

    return altitude


def spec_hum(e, p):
    """Calculates the specific humidity

    Parameters
    ----------
    pressure : float
        air pressure in hPa
    e : float
        vapour pressure in hPa


    Returns
    -------
    spec_hum : float

    Notes
    -----
    The formula is:
    s = ((0.623*e)/(p-0.377*e))*1000
    where M_w/M_tL=0.622
    and 0.378 = 1-0.622"""

    spec_hum = (0.622 * e * 1000) / (p - 0.378 * e)

    return spec_hum


def psychro2e(t_dry, t_wet, p=None):
    """Vapour pressure from dry and wet temperature from Assmann psychrometer
    using Sprung's [1]_ formula (as seen on wikipedia)

    Parameters
    ----------
    t_dry : float or np.array of floats
        dry temperature [deg C]
    t_wet : float or np.array of floats
        wet temperature [deg C]
    p : float or np.array of floats or None, optional
        air pressure [hPa], if None: use the simplified version of the formula,
        default is None

    Returns
    -------
    e : float or np.array of floats
        vapour pressure [hPa]

    Notes
    -----
    If p is None, a simplification is used which can be used below 500m above
    sea level

    References
    ----------
    .. [1] Sprung, A.: Ueber die Bestimmung der Luftfeuchtigkeit mit Hilfe des
        Assmannschen Aspirationspsychrometers, Z. Angew. Meteorol., Das Wetter,
        5 (1888), S. 105?108

    Examples
    --------
    >>> psychro2e(np.array([17.5,18.9]),np.array([12.3,12.1]))
    array([10.82619823,   9.56701424])

    >>> p = np.array([800,800])
    >>> psychro2e(np.array([17.5,18.9]),np.array([12.3,12.1]),p=p)
    array([11.57630294,  10.54309631])
    """
    p = np.array(p)
    if p.any():
        cp = 1005.4  # specific heat capacity of air J/(kg*K)
        mu = 0.622  # molar mass ratio water/air
        # latent heat of vaporization of water [J/kg]
        lam = (
            -0.0000614342 * t_dry**3
            + 0.00158927 * t_dry**2
            - 2.36418 * t_dry
            + 2500.79
        ) * 1000
        gamma = p * cp / (mu * lam)  # hPa/K
    else:
        gamma = 0.67  # hPa/K, simplification, can be used below 500m asl
    return sat_vap_p(t_wet) - gamma * (t_dry - t_wet)


def slope_sat_p(at):
    """slope of the saturation vapor pressure function,
    depends only on air temperature

    Parameters
    ----------
    at : float or numpy.array of floats
        (mean daily) air temperature [deg C]

    Returns
    -------
    slope : float or numpy.array of floats
        slope of the saturation vapor pressure function  [kPa/deg C]

    Notes
    -----
    A polynomial is used to evaluate the slope and is only valid
    for -5 < 'at' > 45 [deg C].

    References
    ----------
    .. [1] Campbell Scientific (1995) application note 4-D: On-Line
        Estimation of Grass Reference Evapotranspiration with the Campbell
        Scientific Automated Weather Station
    """

    slope = (45.3 + 2.97 * at + 0.0549 * at**2 + 0.00223 * at**3) / 1000

    return slope


def esi(at, rh, sw):
    """Environmental stress index (wet bulb globe temperature substitute)

    Parameters
    ----------
    at : float or numpy.array of floats
        air temperature [deg C]
    rh : float or numpy.array of floats
        relative humidity [%]
    sw : float or numpy.array of floats
        solar radiation [W / m²]

    Returns
    -------
    ESI : environmental stress index

    References
    ----------
    .. [1] Moran et al., “An Environmental Stress Index (ESI) as a
       Substitute for the Wet Bulb Globe Temperature (WBGT).”

    """
    return (
        0.63 * at
        - 0.03 * rh
        + 0.002 * sw
        + 0.0054 * (at * rh)
        - 0.073 / (0.1 + sw)
    )


def wet_bulb_stull(at, rh):
    """
    Examples:
    ---------
    >>> wet_bulb_stull(20, 50)
    13.7
    """
    return (
        at * np.arctan(0.151977 * (rh + 8.313659) ** 0.5)
        + np.arctan(at + rh)
        - np.arctan(rh - 1.676331)
        + 0.00391838 * (rh) ** 1.5 * np.arctan(0.023101 * rh)
        - 4.686035
    )


def humidex(at, rh):
    dew = dewpoint(at, rh=rh / 100)
    return at + 0.5555 * (
        6.11 * np.exp(5417.753 * (1 / 273.15 - 1 / (273.15 + dew))) - 10
    )


def apparent_temperature(Ta, rh, ws, Q):
    """Australian Bureau of Meteorology formulation of Steadman (1984)."""
    e = rh / 100 * 6.105 * np.exp(17.27 * Ta / (237.7 + Ta))
    return Ta + 0.348 * e - 0.7 * Q / (ws + 10) - 4.25


def brunner_compound(sti, spi, sequential=False, progress=False):
    """Rank-based hot-dry index.


    Notes
    -----
    Brunner 2021 uses E-GPD for precipitation and a STI index for
    temperature. This implementation just uses empirical ranks.

    References
    ----------
    Brunner, Manuela I., Eric Gilleland, and Andrew W. Wood.
    “Space–Time Dependence of Compound Hot–Dry Events in the United
    States: Assessment Using a Multi-Site Multi-Variable Weather
    Generator.” Earth System Dynamics 12, no. 2 (May 19, 2021):
    621–34. https://doi.org/10.5194/esd-12-621-2021.
    """
    if progress:
        from tqdm import tqdm as progress
    else:

        def progress(x, *args, **kwds):
            return x

    assert len(sti) == len(spi)
    sti_ranks = bottleneck.nanrankdata(sti, axis=-1)
    spi_ranks = bottleneck.nanrankdata(-spi, axis=-1)
    try:
        if sequential:
            raise MemoryError
        bc = np.mean(
            (sti_ranks[..., None, :] <= sti_ranks[..., :, None])
            & (spi_ranks[..., None, :] <= spi_ranks[..., :, None]),
            axis=-1,
        )
        nan_mask = np.isnan(sti_ranks) | np.isnan(spi_ranks)
        bc[nan_mask] = np.nan
    except MemoryError:
        shape_before = sti_ranks.shape
        sti_ranks = sti_ranks.reshape(-1, shape_before[-1])
        spi_ranks = spi_ranks.reshape(-1, shape_before[-1])
        bc = np.full_like(sti_ranks, np.nan)
        for row_i in progress(range(len(sti_ranks)), total=len(sti_ranks)):
            for i, (sti_rank, spi_rank) in enumerate(
                zip(sti_ranks[row_i], spi_ranks[row_i])
            ):
                if np.isnan(sti_rank) or np.isnan(spi_rank):
                    continue
                sti_row = sti_ranks[row_i]
                spi_row = spi_ranks[row_i]
                finite_mask = np.isfinite(sti_row) & np.isfinite(spi_row)
                if np.any(finite_mask):
                    bc[row_i, i] = np.mean(
                        (sti_row <= sti_rank) & (spi_row <= spi_rank),
                        # (sti_row[finite_mask] <= sti_rank)
                        # & (spi_row[finite_mask] <= spi_rank),
                        axis=-1,
                    )
        bc = bc.reshape(shape_before)
    return bc


def _measure_ar(
    varname, distribution, agg_func_name, obs_ar, weeks=1, reference=None
):
    obs_ar = obs_ar.transpose(..., "time").rolling(time=weeks * 7)
    obs_ar = getattr(obs_ar, agg_func_name)()
    if reference is None:
        reference = obs_ar
    else:
        reference = reference.rolling(time=weeks * 7)
        reference = getattr(reference, agg_func_name)()
    months = reference.time.dt.month
    param_names = [param.name for param in distribution._param_info()]
    param_coords = {key: obs_ar.coords[key] for key in obs_ar.dims}
    param_coords["param"] = param_names
    params = xr.DataArray(
        np.empty(obs_ar.shape + (len(param_names),)),
        coords=param_coords,
    )
    for month, monthly_data in reference.groupby(months):
        flat = monthly_data.data.reshape(-1, monthly_data.sizes["time"])
        params_flat = np.array(
            [distribution.fit(row[np.isfinite(row)]) for row in flat]
        )
        params.loc[dict(time=(months == month))] = params_flat.reshape(
            monthly_data.shape[:-1] + (1, len(param_names))
        )
    measure_data = stats.norm.ppf(
        distribution.cdf(
            obs_ar,
            **{
                key: val.data
                for key, val in params.to_dataset("param").items()
            },
        )
    )
    measure_coords = obs_ar.coords.copy()
    measure_coords["variable"] = varname
    # Explicitly specify dims to avoid relying on coord ordering
    # squeeze() removes the 'variable' dimension, so dims are obs_ar.dims without 'variable'
    measure_dims = [dim for dim in obs_ar.dims if dim != "variable"]
    measure_ar = xr.DataArray(
        np.squeeze(measure_data), coords=measure_coords, dims=measure_dims
    )
    return measure_ar


def STI_ds(obs_ds, weeks=1, reference=None):
    sti_ar = _measure_ar(
        "sti",
        stats.norm,
        "mean",
        obs_ds.to_dataarray("variable"),
        weeks=weeks,
        reference=reference,
    )
    return xr.Dataset(dict(sti=sti_ar))


def STI_ar(obs_ar, weeks=1, reference=None):
    return _measure_ar(
        "sti",
        stats.norm,
        "mean",
        obs_ar,
        weeks=weeks,
        reference=reference,
    )


def SPI_ds(obs_ds, weeks=1, reference=None):
    spi_ar = _measure_ar(
        "spi",
        stats.genpareto,
        "sum",
        obs_ds.to_dataarray("variable"),
        weeks=weeks,
        reference=reference,
    )
    return xr.Dataset(dict(spi=spi_ar))


def SPI_ar(obs_ar, weeks=1, reference=None):
    return _measure_ar(
        "spi",
        stats.genpareto,
        "sum",
        obs_ar,
        weeks=weeks,
        reference=reference,
    )


if __name__ == "__main__":
    # doys = varwg.times.datetime2doy(varwg.times.str2datetime(500 * ["2011-09-28T11:27"]))
    # # pot_s_rad(doys)
    # sunshine_pot(doys)
    # from datetime import date
    # sunshine_riseset(date(2018, 6, 1), 8.848, 48.943)

    # import doctest

    # doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)
    pass
